import glob
import math as m

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy import interpolate

# Parameters TODO more general approach for other Electrolysers (here: Stuart HRI)
koh_weight_concentration = 30  # %
anode_surface = 0.03  # m^2
cathode_surface = 0.03  # m^2
anode_thickness = 0.002  # m
cathode_thickness = 0.002  # m
anode_membrane_gap = 0.00125  # m
cathode_membrane_gap = 0.00125  # m
membrane_surface_area = 0.03  # m^2
limiting_current_density = 300000  # A/m^2 (limiting current density)
standard_temperature = 25  # °C (Room Temperature)
standard_temperature = standard_temperature + 273.15  # K

__FARADAY_CONST: float = 96485.3321  # As/mol
__IDEAL_GAS_CONST: float = 8.314462  # J/(mol K)


def calculate_bubble_rate_coverage(current: float, temperature: float) -> float:
    # TODO Check variation in Formula from Olivier Paper (2 instead of 0.3)
    bubble_rate_coverage = (-97.25 + 182 * (temperature / standard_temperature) - 84 * (
                temperature / standard_temperature) ** 2) * (
                                   (current / cathode_surface) / limiting_current_density) ** 0.3

    return bubble_rate_coverage


def calculate_anode_activation_current(activation_overvoltage: float, temperature: float) -> float:
    alpha_a = 0.0675 + 0.00095 * temperature
    exchange_current_density = 30.4 - 0.206 * temperature + 0.00035 * temperature ** 2  # ma/cm2 ?
    exchange_current_density = exchange_current_density / 1000.0  # A/cm2

    factor = (2 * __FARADAY_CONST * activation_overvoltage) / (
                __IDEAL_GAS_CONST * temperature)
    current_density_anode = exchange_current_density * (m.exp(alpha_a * factor) - m.exp(-(1 - alpha_a) * factor))

    return current_density_anode


def calculate_cathode_activation_current(activation_overvoltage: float, temperature: float) -> float:
    alpha_c = 0.1175 + 0.00095 * temperature
    exchange_current_density = 13.72491 - 0.09055 * temperature + 0.09055 * temperature ** 2
    exchange_current_density = exchange_current_density / 1000.0  # A/cm2

    factor = (2 * __FARADAY_CONST * activation_overvoltage) / (
                __IDEAL_GAS_CONST * temperature)
    current_density_cathode = exchange_current_density * (m.exp(alpha_c * factor) - m.exp(-(1 - alpha_c) * factor))

    return current_density_cathode


def generate_activation_overvoltage_lookup():
    # Input Variables
    # temperatures = [22.0, 30.0, 40.0, 50.0, 60.0, 70.0, 80.0, 90.0, 100.0, 110.0, 120.0, 130.0, 140.0, 150.0, 160.0]  # °C
    temperatures = np.linspace(22.0, 160.0, 100)
    activation_overvoltages = np.linspace(0, 1, 5000)
    current_densities = np.linspace(0, 0.5, 1000)

    lookup_temperatures = []
    lookup_v_activation_anode = []
    lookup_v_activation_cathode = []
    lookup_current_densities = []

    for t in range(len(temperatures)):
        current_densities_cathode = []
        current_densities_anode = []

        temperature = temperatures[t]
        temperature = temperature + 273.15  # K

        for i in range(len(activation_overvoltages)):
            current_densities_anode.append(calculate_anode_activation_current(activation_overvoltages[i], temperature))
            current_densities_cathode.append(
                calculate_cathode_activation_current(activation_overvoltages[i], temperature))

        interpolation_cathode = interpolate.interp1d(current_densities_cathode, activation_overvoltages,
                                                     fill_value='extrapolate')
        interpolation_anode = interpolate.interp1d(current_densities_anode, activation_overvoltages,
                                                   fill_value='extrapolate')

        for j in range(len(current_densities)):
            v_activation_cathode = interpolation_cathode(current_densities[j])
            v_activation_anode = interpolation_anode(current_densities[j])
            lookup_v_activation_anode.append(v_activation_anode)
            lookup_v_activation_cathode.append(v_activation_cathode)
            lookup_temperatures.append(temperature - 273.15)
            lookup_current_densities.append(current_densities[j])

    lookup_v_activation_cathode = np.array(lookup_v_activation_cathode)
    lookup_v_activation_anode = np.array(lookup_v_activation_anode)
    lookup_temperatures = np.array(lookup_temperatures)
    lookup_current_densities = np.array(lookup_current_densities)

    # Create lookup table for alkaline electrolyser cell activation overpotential from current densitiy
    dict1 = {'temperature in °C': lookup_temperatures, 'current density in A/cm2': lookup_current_densities,
             'cathode activation voltage in V': lookup_v_activation_cathode,
             'anode activation voltage in V': lookup_v_activation_anode}
    df1 = pd.DataFrame(dict1)
    df1.to_csv('activation_overpotential_lookup_alkaline.csv', index=False)


def display_example(temperature: float, compare: bool):
    temperature = temperature + 273.15
    activation_overvoltages = np.linspace(0, 1, 10000)

    current_densities_cathode = []
    current_densities_anode = []

    for j in range(len(activation_overvoltages)):
        current_densities_anode.append(calculate_anode_activation_current(activation_overvoltages[j], temperature))
        current_densities_cathode.append(calculate_cathode_activation_current(activation_overvoltages[j], temperature))

    current_densities = np.linspace(0, 0.5, 1000)
    interpolation_cathode = interpolate.interp1d(current_densities_cathode, activation_overvoltages,
                                                 fill_value='extrapolate')
    interpolation_anode = interpolate.interp1d(current_densities_anode, activation_overvoltages,
                                               fill_value='extrapolate')

    v_activation_total = []
    for j1 in range(len(current_densities)):
        current = current_densities[j1] * anode_surface * 10000
        bubble_rate_coverage = calculate_bubble_rate_coverage(current, temperature)
        electrode_surface_nom = anode_surface * (1 - bubble_rate_coverage)
        current_density = current / (electrode_surface_nom * 10000)  # A/cm2
        v_activation_cathode = interpolation_cathode(current_density)
        v_activation_anode = interpolation_anode(current_density)
        v_activation_total.append(v_activation_cathode + v_activation_anode)

    plt.subplot(111)
    plt.ylabel('Activation overpotential in V')
    plt.xlabel('Current density in A/cm2')
    plt.title('Activation overpotential')
    plt.grid(True)
    plt.plot(current_densities, v_activation_total)
    if compare:
        # Compare activation overpotential curves from source paper
        legend = []
        legend.append('Calculated activation overpotential at ' + str(temperature - 273.15) + '°C')

        activation_curve_path = 'ui_curves/activation_overpotential_alkaline\\'
        u_i_files = glob.glob(activation_curve_path + '*.csv')
        for curve in u_i_files:
            excelfile = pd.read_csv(curve)
            file_current_densities = excelfile.iloc[:, 0]
            file_activation_voltages = excelfile.iloc[:, 1]
            plt.plot(file_current_densities.values, file_activation_voltages.values, linewidth=2)
            legend.append(curve[len(activation_curve_path):len(curve) - 4])

        plt.legend(legend)

    plt.show()


"""Command Section"""
# generate_activation_overvoltage_lookup()
# display_example(25, compare=True)


from configparser import ConfigParser

import numpy as np
import pytest

from simses.commons.config.data.battery import BatteryDataConfig
from simses.commons.config.simulation.battery import BatteryConfig
from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.commons.utils.utilities import all_non_abstract_subclasses_of
from simses.technology.lithium_ion.cell.generic import GenericCell
from simses.technology.lithium_ion.cell.isea import IseaCellType
from simses.technology.lithium_ion.cell.lfp_sony import SonyLFP
from simses.technology.lithium_ion.cell.lmo_daimler import DaimlerLMO
from simses.technology.lithium_ion.cell.lto_lmo import LTOLMO
from simses.technology.lithium_ion.cell.lto_nmc import LTONMC
from simses.technology.lithium_ion.cell.nca_panasonic_ncr import PanasonicNCA
from simses.technology.lithium_ion.cell.nmc_molicel import MolicelNMC
from simses.technology.lithium_ion.cell.nmc_samsung78Ah import Samsung78AhNMC
from simses.technology.lithium_ion.cell.nmc_samsung94Ah import Samsung94AhNMC
from simses.technology.lithium_ion.cell.nmc_samsung94Ah_hybrid import Samsung94AhNMCHybrid
from simses.technology.lithium_ion.cell.nmc_sanyo_ur18650e import SanyoNMC
from simses.technology.lithium_ion.cell.type import CellType


class TestCellType:
    # Single cell tests
    lithium_ion_state: LithiumIonState = LithiumIonState(0,0)
    step = 10
    test_values_soc = np.arange(0, 1.01, 1 / step)

    def make_lithium_ion_cell(self, lithium_ion_subclass):
        voltage = 0.1
        capacity = 0.1
        soh = 1.0
        config: ConfigParser = ConfigParser()
        if lithium_ion_subclass.__name__ in [GenericCell.__name__, Samsung78AhNMC.__name__]:
            return lithium_ion_subclass(voltage, capacity, soh, BatteryConfig(config=config))
        return lithium_ion_subclass(voltage, capacity, soh, BatteryConfig(config=config), BatteryDataConfig())

    @pytest.fixture(scope="module") # module: gets only one instance
    def lithium_ion_subclass_list(self):
        return all_non_abstract_subclasses_of(CellType)

    @pytest.mark.parametrize("soc", test_values_soc)
    def test_ocv(self, soc, lithium_ion_subclass_list):
        for lithium_ion_subclass in lithium_ion_subclass_list:
            if lithium_ion_subclass.__name__ == IseaCellType.__name__:
                continue
            uut = self.make_lithium_ion_cell(lithium_ion_subclass)
            self.lithium_ion_state.soc = soc
            # all ocv values must be between the cell minimal voltage and maximum voltage
            assert uut.get_open_circuit_voltage(self.lithium_ion_state) >= uut.get_min_voltage()
            assert uut.get_open_circuit_voltage(self.lithium_ion_state) <= uut.get_max_voltage()



from simses.system.dc_coupling.dc_coupler import DcCoupling
from simses.system.dc_coupling.generation.no_dc_generation import NoDcGeneration
from simses.system.dc_coupling.load.no_dc_load import NoDcLoad


class NoDcCoupling(DcCoupling):

    def __init__(self):
        super().__init__(NoDcLoad(), NoDcGeneration())



from configparser import ConfigParser

from simses.commons.config.simulation.simulation_config import SimulationConfig


class BatteryConfig(SimulationConfig):
    """
    Battery specific configs
    """

    SECTION: str = 'BATTERY'
    START_SOC: str = 'START_SOC'
    MIN_SOC: str = 'MIN_SOC'
    MAX_SOC: str = 'MAX_SOC'
    EOL:str = 'EOL'
    START_SOH: str = 'START_SOH'
    START_SOH_SHARE: str = 'START_SOH_SHARE'
    EXACT_SIZE: str = 'EXACT_SIZE'
    CELL_SERIAL_SCALE: str = 'CELL_SERIAL_SCALE'
    CELL_PARALLEL_SCALE: str = 'CELL_PARALLEL_SCALE'
    DEGRADATION_MODEL_NUMBER: str = 'DEGRADATION_MODEL_NUMBER'
    CONSIDER_VOLTAGE_LIMIT: str = 'CONSIDER_VOLTAGE_LIMIT'
    MULTI_MODEL_PARAMETERS: str = 'MULTI_MODEL_PARAMETERS'
    CALENDAR_LIFETIME: str = 'CALENDAR_LIFETIME'
    CYCLE_LIFETIME: str = 'CYCLE_LIFETIME'

    def __init__(self, config: ConfigParser, path: str = None):
        super().__init__(path, config)

    @property
    def soc(self) -> float:
        """
        Minimum SOC (0-1)

        Returns
        -------
        float:
            Returns the start soc from data_config file

        """
        return float(self.get_property(self.SECTION, self.START_SOC))

    @property
    def min_soc(self) -> float:
        """
        Minimum SOC (0-1)

        Returns
        -------
        float:
            Returns the minimum soc from data_config file

        """
        return float(self.get_property(self.SECTION, self.MIN_SOC))

    @property
    def max_soc(self) -> float:
        """
        Maximum SOC (0-1)

        Returns
        -------
        float:
            Returns the maximum soc from data_config file

        """
        return float(self.get_property(self.SECTION, self.MAX_SOC))

    @property
    def eol(self) -> float:
        """
        End of Life criteria (0-1)

        Returns
        -------
        float:
            Returns EOL criteria in % from data_config file

        """
        return float(self.get_property(self.SECTION, self.EOL))

    @property
    def start_soh(self) -> float:
        """
        End of Life criteria (0-1)

        Returns
        -------
        float:
            Returns start SOH from data_config file

        """
        return float(self.get_property(self.SECTION, self.START_SOH))

    @property
    def exact_size(self) -> bool:
        """Returns selection for exact sizing True/False"""
        try:
            return self.get_property(self.SECTION, self.EXACT_SIZE) in ['True', None]
        except (KeyError, TypeError):
            return True

    @property
    def start_soh_share(self) -> float:
        """
        Share of start SOH between calendar and cyclic degradation for both, capacity decrease and resistance increase

        Returns
        -------
        float:
            Returns start SOH share in p.u.

        """
        return float(self.get_property(self.SECTION, self.START_SOH_SHARE))

    @property
    def degradation_model_number(self) -> int:
        """
        Number / identifier for selected degradation model for degradation models that read parameters from CSV files
        (e.g. SonyLFP).

        Returns
        -------
        int:
            Returns the degradation model parameter number that is to be read from the CSV.
            Returns 1 if not specified in config.

        """
        try:
            return int(self.get_property(self.SECTION, self.DEGRADATION_MODEL_NUMBER))
        except (KeyError, TypeError):
            return 1

    @property
    def multi_model_parameters(self) -> str:
        """
        Parameters for multi model degradation models. Under development.

        Returns
        -------
        str:
            Comma seperated parameters in no particular order

        """
        return str(self.get_property(self.SECTION, self.MULTI_MODEL_PARAMETERS))

    @property
    def serial_scale(self) -> float:
        """Returns a linear scaling factor of cell in order to simulate a serial lithium_ion connection"""
        return float(self.get_property(self.SECTION, self.CELL_SERIAL_SCALE))

    @property
    def parallel_scale(self) -> float:
        """Returns a linear scaling factor of cell in order to simulate a parallel lithium_ion connection"""
        return float(self.get_property(self.SECTION, self.CELL_PARALLEL_SCALE))

    @property
    def max_equivalent_full_cycles(self) -> float:
        """Returns a number of maximum equivalent full cycles for GenericCell degradation"""
        try:
            return float(self.get_property(self.SECTION, self.CYCLE_LIFETIME))
        except TypeError:
            return 2000.0

    @property
    def max_calendar_lifetime(self) -> float:
        """Returns a maximum calendar lifetime for GenericCell degradation in years"""
        try:
            return float(self.get_property(self.SECTION, self.CALENDAR_LIFETIME))
        except TypeError:
            return 5.

    @property
    def consider_voltage_limit(self) -> bool:
        """Returns if current should be limited by charge / discharge end voltage, default: True"""
        try:
            return self.get_property(self.SECTION, self.CONSIDER_VOLTAGE_LIMIT) in ['True', None]
        except (KeyError, TypeError):
            return True



from simses.commons.config.data.battery import BatteryDataConfig
from simses.commons.config.simulation.battery import BatteryConfig
from simses.commons.cycle_detection.cycle_detector import CycleDetector
from simses.technology.lithium_ion.cell.type import CellType
from simses.technology.lithium_ion.degradation.calendar.lfp_sony import \
    SonyLFPCalendarDegradationModel
from simses.technology.lithium_ion.degradation.cyclic.lfp_sony import \
    SonyLFPCyclicDegradationModel
from simses.technology.lithium_ion.degradation.degradation_model import DegradationModel


class SonyLFPDegradationModel(DegradationModel):

    def __init__(self, cell_type: CellType, cycle_detector: CycleDetector, battery_config: BatteryConfig,
                 battery_data_config: BatteryDataConfig):
        super().__init__(cell_type,
                         SonyLFPCyclicDegradationModel(cell_type, cycle_detector, battery_data_config, battery_config),
                         SonyLFPCalendarDegradationModel(cell_type, battery_data_config, battery_config),
                         cycle_detector, battery_config, initial_degradation_possible=True)



from configparser import ConfigParser

from simses.commons.config.data.data_config import DataConfig


class PowerElectronicsConfig(DataConfig):
    """class top read Power Electronics data path"""

    def __init__(self, path: str = None, config: ConfigParser = None):
        super().__init__(path, config)
        self.__section: str = 'POWER_ELECTRONICS'

    @property
    def acdc_converter_data(self) -> str:
        """Returns directory of acdc converter data files"""
        return self.get_data_path(self.get_property(self.__section, 'ACDC_CONVERTER_DATA'))

    @property
    def sinamics_efficiency_file(self) -> str:
        """Returns filename for Siemens S120 converter"""
        return self.acdc_converter_data + self.get_property(self.__section, 'SINAMICS_EFFICIENCY_FILE')

    @property
    def aixcontrol_efficiency_file(self) -> str:
        """Returns filename for AixControl GmbH converter"""
        return self.acdc_converter_data + self.get_property(self.__section, 'AIXCONTROL_EFFICIENCY_FILE')

    @property
    def dcdc_converter_dir(self) -> str:
        """Returns directory of acdc converter data files"""
        return self.get_data_path(self.get_property(self.__section, 'DCDC_CONVERTER_DIR'))

    @property
    def pgs_efficiency_file(self) -> str:
        """Returns filename for Siemens S120 converter"""
        return self.dcdc_converter_dir + self.get_property(self.__section, 'PGS_EFFICIENCY_FILE')



from abc import ABC, abstractmethod

from simses.system.auxiliary.auxiliary import Auxiliary


class DcGeneration(ABC):

    def __init__(self):
        super().__init__()

    @abstractmethod
    def get_power(self) -> float:
        """

        Returns
        -------
        float:
            DC load power in W
        """
        pass

    @abstractmethod
    def calculate_power(self, time: float) -> None:
        """
        Calculates power for next timestep

        Parameters
        ----------
        time :
            timestamp in s
        """
        pass

    @abstractmethod
    def get_auxiliaries(self) -> [Auxiliary]:
        pass

    @abstractmethod
    def close(self):
        pass



from simses.commons.constants import Hydrogen
from simses.system.auxiliary.compression.compressor import Compressor


class HydrogenIsentropCompressor(Compressor):

    def __init__(self, compressor_eta = 0.95):
        super().__init__()
        self.__compressor_eta = compressor_eta
        self.__compression_power = 0  # W
        self.__R = Hydrogen.IDEAL_GAS_CONST

    def calculate_compression_power(self, hydrogen_flow_out:float, pressure_1: float, pressure_2: float, temperature: float) -> None:
        real_gas_faktor = Hydrogen.realgas_factor(pressure_1, pressure_2)
        isentropic_exponent = Hydrogen.isentropic_exponent(pressure_1, pressure_2)
        prefactor = isentropic_exponent / (isentropic_exponent - 1)
        self.__compression_power = 1.0 / self.__compressor_eta * prefactor * self.__R * (temperature + 273.15) * \
                                   real_gas_faktor * (( pressure_2 / pressure_1) ** (1 / prefactor) - 1) * hydrogen_flow_out

    def get_compression_power(self) -> float:
        return self.__compression_power



import scipy.integrate as integrate
import pandas as pd
from simses.commons.cycle_detection.cycle_detector import CycleDetector
from simses.commons.log import Logger
from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.technology.lithium_ion.cell.type import CellType
from simses.technology.lithium_ion.degradation.cyclic.cyclic_degradation import \
    CyclicDegradationModel
from simses.commons.config.data.battery import BatteryDataConfig
from simses.commons.config.simulation.battery import BatteryConfig


class SonyLFPCyclicDegradationModel(CyclicDegradationModel):

    def __init__(self, cell_type: CellType, cycle_detector: CycleDetector, battery_data_config: BatteryDataConfig,
                 battery_config: BatteryConfig):
        super().__init__(cell_type, cycle_detector)
        self.__log: Logger = Logger(type(self).__name__)
        self.__capacity_loss = 0
        self.__capacity_loss_cyclic = cell_type.get_cyclic_capacity_loss_start()
        self.__resistance_increase = 0
        self.__initial_capacity = self._cell.get_nominal_capacity()

        # Source SONY_US26650FTC1_Product Specification and Naumann, Maik, Franz Spingler, and Andreas Jossen.
        # "Analysis and modeling of cycle aging of a commercial LiFePO4/graphite cell."
        # Journal of Power Sources 451 (2020): 227666.
        # DOI: https://doi.org/10.1016/j.jpowsour.2019.227666

        # # original parameters
        # self.__A_QLOSS = 0.0630  # constant
        # self.__B_QLOSS = 0.0971  # constant
        # self.__C_QLOSS = 4.0253  # constant
        # self.__D_QLOSS = 1.0923  # constant
        #
        # self.__A_RINC = -0.0020  # constant
        # self.__B_RINC = 0.0021  # constant
        # self.__C_RINC = 6.8477  # constant
        # self.__D_RINC = 0.91882  # constant

        # read in parameters from CSV
        model_number = battery_config.degradation_model_number
        values_cap_model = pd.read_csv(battery_data_config.lfp_sony_degradation_capacity_file,
                                       skiprows=lambda x: x not in [0, model_number]).values.tolist()[0]
        values_res_model = pd.read_csv(battery_data_config.lfp_sony_degradation_resistance_file,
                                       skiprows=lambda x: x not in [0, model_number]).values.tolist()[0]
        # print("\nUsing the following cyc degradation model: \tNumber: " + str(model_number) + "\tIdentifier: " +
        #       values_cap_model[0])
        self.__A_QLOSS = values_cap_model[5]  # constant
        self.__B_QLOSS = values_cap_model[6]  # constant
        self.__C_QLOSS = values_cap_model[7]  # constant
        self.__D_QLOSS = values_cap_model[8]  # constant

        self.__A_RINC = values_res_model[5]  # constant
        self.__B_RINC = values_res_model[6]  # constant
        self.__C_RINC = values_res_model[7]  # constant
        self.__D_RINC = values_res_model[8]  # constant

    def calculate_degradation(self, battery_state: LithiumIonState) -> None:
        crate: float = self._cycle_detector.get_crate() * 3600  # in 1 / s -> *3600 -> in 1/h
        doc: float = self._cycle_detector.get_depth_of_cycle()  # in pu
        delta_fec: float = self._cycle_detector.get_delta_full_equivalent_cycle()  # in pu
        qloss: float = (self.__initial_capacity - battery_state.capacity / battery_state.nominal_voltage) / self.__initial_capacity  # pu

        # calculate stress factor dependent coefficients
        k_c_rate_qloss = self.__A_QLOSS * crate + self.__B_QLOSS
        k_doc_qloss = self.__C_QLOSS * (doc - 0.6)**3 + self.__D_QLOSS

        # # calculate capacity loss based on virtual FEC and past total degradation.
        # virtual_fec: float = (qloss * 100 / (k_c_rate_qloss * k_doc_qloss))**2
        # capacity_loss = k_c_rate_qloss * k_doc_qloss * (virtual_fec + delta_fec)**0.5 / 100  # total cyc. qloss in p.u.
        # capacity_loss -= qloss  # relative qloss in pu in current timestep

        # calculate capacity loss based on virtual FEC and past cyclic degradation.
        virtual_fec: float = (self.__capacity_loss_cyclic * 100 / (k_c_rate_qloss * k_doc_qloss))**2
        capacity_loss = k_c_rate_qloss * k_doc_qloss * (virtual_fec + delta_fec)**0.5 / 100  # total cyc. qloss in p.u.
        capacity_loss -= self.__capacity_loss_cyclic  # relative qloss in pu in current timestep
        self.__capacity_loss_cyclic += capacity_loss

        self.__capacity_loss = capacity_loss * self.__initial_capacity  # Ah

    def calculate_resistance_increase(self, battery_state: LithiumIonState) -> None:
        crate: float = self._cycle_detector.get_crate() * 3600  # in 1 / s -> *3600 -> in 1/h
        doc: float = self._cycle_detector.get_depth_of_cycle()  # in pu
        delta_fec: float = self._cycle_detector.get_delta_full_equivalent_cycle()  # in pu

        # calculate stress factor dependent coefficients
        k_c_rate_rinc = self.__A_RINC * crate + self.__B_RINC
        k_doc_rinc = self.__C_RINC * (doc - 0.5)**3 + self.__D_RINC

        # calculate resistance increase
        resistance_increase = k_c_rate_rinc * k_doc_rinc * delta_fec
        self.__resistance_increase = resistance_increase / 100  # in pu

    def get_degradation(self) -> float:
        capacity_loss = self.__capacity_loss
        self.__capacity_loss = 0    # Set value to 0, because cyclic losses are not calculated in each step
        return capacity_loss

    def get_resistance_increase(self) -> float:
        resistance_increase = self.__resistance_increase
        self.__resistance_increase = 0  # Set value to 0, because cyclic losses are not calculated in each step
        return resistance_increase

    def reset(self, lithium_ion_state: LithiumIonState) -> None:
        self.__capacity_loss = 0
        self.__resistance_increase = 0

    def close(self) -> None:
        self.__log.close()



from ast import literal_eval
from configparser import ConfigParser
from simses.commons.state.parameters import SystemParameters


class SystemParametersReader:
    """
    SystemParametersReader provides a third interface to SimSES for external programs which need to access variables
    whose values are known only after the SimSES object has been instantiated.
    This class reads the SystemParameters.txt file generated by the SimSES constructor after instantiation, and provides
    specific values contained therein.
    An example of one such variable: the number of containers

    """

    def __init__(self, path: str, name: str = None):
        config: ConfigParser = ConfigParser()
        if name:
            file_path = path + name + '/' + SystemParameters.get_file_name()
        else:
            file_path = path + SystemParameters.get_file_name()
        config.read(file_path)
        self.__system_parameters: dict = literal_eval(config.get(SystemParameters.SECTION, SystemParameters.PARAMETERS))

    def get_all_storage_systems_ac(self) -> [dict]:  # TODO consider if this should be a private method
        """
        Obtaining all StorageSystemAC objects in system
        :return: System parameters of all StorageSystemAC objects created as a list of dict
        """
        return self.__system_parameters[SystemParameters.SUBSYSTEM]

    def get_number_of_storage_systems_ac(self) -> int:
        """
        Obtaining number of StorageSystemAC objects in system
        :return: number of StorageSystemAC objects created as int
        """
        return len(self.get_all_storage_systems_ac())

    def get_storage_system_ac_id_for(self, system: dict) -> int:
        return int(system[SystemParameters.ID])

    def get_storage_system_ac(self, id: int) -> dict:  # TODO consider if this should be a private method
        """
        Querying a StorageSystemAC object by ID
        :param id: ID of the StorageSystemAC object to be queried as int
        :return: attributes of the queried StorageSystemAC object as a dict
        """
        systems: dict = self.get_all_storage_systems_ac()
        for system in systems:
            if self.get_storage_system_ac_id_for(system) == id:
                return system
        raise Exception('Requested storage system AC not available.')

    def get_container_type_for(self, id: int) -> str:
        """
        Query the Housing object type for specified StorageSystemAC object
        :param id: ID of the StorageSystemAC object to be queried as int
        :return: name of the container Housing object type as str
        """
        system: dict = self.get_storage_system_ac(id)
        type: str = system[SystemParameters.CONTAINER_TYPE]
        return type

    def get_number_of_containers_for(self, id: int) -> int:
        """
        Query the number of container for specified StorageSystemAC object
        :param id: ID of the StorageSystemAC object to be queried as int
        :return: number of containers in the StorageSystemAC object as int
        """
        system: dict = self.get_storage_system_ac(id)
        number_containers = system[SystemParameters.CONTAINER_NUMBER]
        return number_containers

    def get_total_containers(self) -> int:
        """
        Obtain the total number of containers for all StorageSystemAC objects in system
        :return: total number of containers as int
        """
        total_number_containers: int = 0
        systems: [dict] = self.get_all_storage_systems_ac()
        for system in systems:
            system_id: int = self.get_storage_system_ac_id_for(system)
            if self.get_number_of_containers_for(system_id) > 0:
                print(system[SystemParameters.SYSTEM] + ' ' + str(system_id) +
                      ' has ' + str(self.get_number_of_containers_for(system_id)) + ' container(s) of type '
                      + self.get_container_type_for(system_id))
                total_number_containers += self.get_number_of_containers_for(system_id)
            else:
                print(system[SystemParameters.SYSTEM] + ' ' + str(system_id) +
                      ' has ' + self.get_container_type_for(system_id))
        return total_number_containers

    def get_acdc_converter_type_for(self, id: int) -> str:
        """
        Query the ACDC Converter object type for specified StorageSystemAC object
        :param id: ID of the StorageSystemAC object to be queried as int
        :return: name of the ACDC Converter object type as str
        """
        system: dict = self.get_storage_system_ac(id)
        type: str = system[SystemParameters.ACDC_CONVERTER]
        return type

    def get_all_storage_systems_dc_for(self, id: int) -> [dict]:  # TODO consider if this should be a private method
        """
        Obtaining all StorageSystemDC objects in specified StorageSystemAC
        :param id: id of StorageSystemAC
        :return: System parameters of all StorageSystemDC objects as a list of dict
        """
        system: dict= self.get_storage_system_ac(id)
        dc_systems: [dict] = system[SystemParameters.SUBSYSTEM]
        return dc_systems

    def get_number_of_storage_systems_dc(self, id: int) -> int:
        """
        Obtain number of StorageSystemDC objects in specified StorageSystemAC
        :param id: ID of the StorageSystemAC object to be queried as int
        :return: number of StorageSystemDC objects in specified StorageSystemAC as int
        """
        return len(self.get_all_storage_systems_dc_for(id))

    def get_storage_system_dc_str_id_for(self, system: dict) -> str:
        """
        Returns the ID of the provided StorageSystemDC as str with the ID of StorageSystemAC
        :param system: StorageSystemDC as dict
        :return: ID of the provided StorageSystemDC as str
        """
        return str(system[SystemParameters.ID])

    def get_storage_system_dc_id_for(self, system: dict) -> int:
        """
        Returns the ID of the provided StorageSystemDC as int without the ID of StorageSystemAC
        :param system: StorageSystemDC as dict
        :return: ID of the provided StorageSystemDC as int
        """
        str_id = self.get_storage_system_dc_str_id_for(system)
        id = int(float(str_id.split('.')[1]))
        return id

    def get_storage_system_dc(self, ac_id: int, dc_id: int) -> dict:  # TODO consider if this should be a private method
        """
        Query a StorageSystemDC object by StorageSystemAC ID and StorageSystemDC ID
        :param dc_id_str: ID of the StorageSystemDC object to be queried as str (also contains the StorageSystemAC ID)
        :param ac_id: ID of the StorageSystemAC object to be queried as int
        :param dc_id: ID of the StorageSystemDC object to be queried as int
        :return: attributes of the queried StorageSystemDC object as dict
        """
        systems: dict = self.get_all_storage_systems_dc_for(ac_id)
        system_id = str(ac_id) + '.' + str(dc_id)
        for system in systems:
            if self.get_storage_system_dc_str_id_for(system) == system_id:
                return system
        raise Exception('Requested storage system DC not available.')

    def get_dcdc_converter_type_for(self, ac_id: int, dc_id: int) -> str:
        """
        Query the DCDC Converter object type for specified StorageSystemDC object
        :param ac_id: ID of the StorageSystemAC object to be queried as int
        :param dc_id: ID of the StorageSystemDC object to be queried as int
        :return: name of the DCDC Converter object type as str
        """
        system: dict = self.get_storage_system_dc(ac_id, dc_id)
        type: str = system[SystemParameters.DCDC_CONVERTER]
        return type

    def get_storage_technology_for(self, ac_id: int, dc_id: int) -> str:
        """
        Query the Storage Technology object type for specified StorageSystemDC object
        :param ac_id: ID of the StorageSystemAC object to be queried as int
        :param dc_id: ID of the StorageSystemDC object to be queried as int
        :return: name of the Storage Technology object type as str
        """
        system: dict = self.get_storage_system_dc(ac_id, dc_id)
        type: str = system[SystemParameters.STORAGE_TECHNOLOGY]
        return type

    def get_cell_type_for(self, ac_id: int, dc_id: int) -> str:
        """
        Query the Cell type for specified StorageSystemDC object
        :param ac_id: ID of the StorageSystemAC object to be queried as int
        :param dc_id: ID of the StorageSystemDC object to be queried as int
        :return: name of the Cell type as str
        """
        system: dict = self.get_storage_system_dc(ac_id, dc_id)
        type: str = system[SystemParameters.CELL_TYPE]
        return type

    def get_battery_circuit_for(self, ac_id: int, dc_id: int) -> str:
        """
        Query the Battery Circuit for specified StorageSystemDC object
        :param ac_id: ID of the StorageSystemAC object to be queried as int
        :param dc_id: ID of the StorageSystemDC object to be queried as int
        :return: Battery Circuit as str
        """
        system: dict = self.get_storage_system_dc(ac_id, dc_id)
        circuit: str = system[SystemParameters.BATTERY_CIRCUIT]
        return circuit

    def get_auxiliaries_for(self, id: int) -> [str]:
        """
        Query the Auxiliary Components for specified StorageSystemDC object.
        :param id: ID of the StorageSystemAC object to be queried as int
        :return: list of auxiliary components as list of str
        """
        system: dict= self.get_storage_system_ac(id)
        auxiliaries: [str] = system[SystemParameters.AUXILIARIES]
        return auxiliaries

    def get_all_acdc_converters(self) -> [str]:  # TODO consider if this should be a private method
        """
        Query all ACDC Converter objects
        :return: Names of all DCDC Converter objects as list of str
        """
        systems: dict = self.get_all_storage_systems_ac()
        acdc_converters: [str] = list()
        for system in systems:
            acdc_converters.append(self.get_acdc_converter_type_for(self.get_storage_system_ac_id_for(system)))
        return acdc_converters

    def get_all_auxiliaries(self) -> [list]:  # TODO consider if this should be a private method
        """
        Query all Auxiliary System objects in specified StorageSystemAC
        :return: Names of all Auxiliary System objects in specified StorageSystemAC as list of str
        """
        systems: dict = self.get_all_storage_systems_ac()
        auxiliaries: [list] = list()
        for system in systems:
            auxiliaries.append(self.get_auxiliaries_for(self.get_storage_system_ac_id_for(system)))
        return auxiliaries

    def get_all_dcdc_converters(self, ac_id: int) -> [str]:  # TODO consider if this should be a private method
        """
        Query all DCDC Converter objects in specified StorageSystemAC
        :param ac_id: ID of StorageSystemAC
        :return: Names of all DCDC Converter objects in specified StorageSystemAC as list of str
        """
        dc_systems = self.get_all_storage_systems_dc_for(ac_id)
        dcdc_converters: [str] = list()
        for dc_system in dc_systems:
            dc_id: int = self.get_storage_system_dc_id_for(dc_system)
            dcdc_converters.append(self.get_dcdc_converter_type_for(ac_id, dc_id))
        return dcdc_converters










from simses.commons.log import Logger
from simses.commons.utils.utilities import check
from simses.system.power_electronics.acdc_converter.abstract_acdc_converter import AcDcConverter


class FixEfficiencyAcDcConverter(AcDcConverter):

    __VOLUMETRIC_POWER_DENSITY = 143 * 1e6  # W / m3
    __GRAVIMETRIC_POWER_DENSITY = 17000  # W/kg
    __SPECIFIC_SURFACE_AREA = 0.0001  # in m2 / W  # TODO add exact values
    # Exemplary value from:
    # (https://www.iisb.fraunhofer.de/en/research_areas/vehicle_electronics/dcdc_converters/High_Power_Density.html)
    # ( https://www.apcuk.co.uk/app/uploads/2018/02/PE_Full_Pack.pdf )

    def __init__(self, max_power: float, efficiency: float = 0.95):
        super().__init__(max_power)
        self.__EFFICIENCY = efficiency
        self.__log: Logger = Logger(type(self).__name__)
        self.__log.info('efficiency is fixed at ' + str(efficiency))

    def to_ac(self, power: float, voltage: float) -> float:
        check(power)
        res: float = 0.0
        if power < 0.0:
            res = power / self.__EFFICIENCY
        return res

    def to_dc(self, power: float, voltage: float) -> float:
        check(power)
        res: float = 0.0
        if power > 0.0:
            res = power * self.__EFFICIENCY
        return res

    def to_dc_reverse(self, dc_power: float, voltage: float) -> float:
        check(dc_power)
        res: float = 0.0
        if dc_power > 0.0:
            res = dc_power / self.__EFFICIENCY
        return res

    def to_ac_reverse(self, dc_power: float, voltage: float) -> float:
        check(dc_power)
        res: float = 0.0
        if dc_power < 0.0:
            res = dc_power * self.__EFFICIENCY
        return res

    @property
    def volume(self) -> float:
        return self.max_power / self.__VOLUMETRIC_POWER_DENSITY

    @property
    def mass(self):
        return self.max_power / self.__GRAVIMETRIC_POWER_DENSITY

    @property
    def surface_area(self) -> float:
        return self.max_power * self.__SPECIFIC_SURFACE_AREA

    @classmethod
    def create_instance(cls, max_power: float, power_electronics_config=None):
        return FixEfficiencyAcDcConverter(max_power)

    def close(self) -> None:
        pass



from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.config.simulation.profile import ProfileConfig
from simses.commons.profile.technical.soc import SocProfile
from simses.commons.state.energy_management import EnergyManagementState
from simses.commons.state.system import SystemState
from simses.logic.energy_management.strategy.operation_priority import OperationPriority
from simses.logic.energy_management.strategy.operation_strategy import OperationStrategy


class SocFollower(OperationStrategy):
    """
    SOC Follower is a basic operation strategy which converts a given soc profile to a power profile and forward it
    to the storage system.
    """
    def __init__(self, config: GeneralSimulationConfig, profile_config: ProfileConfig):
        super().__init__(OperationPriority.MEDIUM)
        self.__soc_profile = SocProfile(config, profile_config)
        self.__timestep = config.timestep
        self.__Wh_to_Ws = 3600

    def next(self, time: float, system_state: SystemState, power: float = 0) -> float:
        soc_dif = self.__soc_profile.next(time) - system_state.soc
        capacity_dif = soc_dif * system_state.capacity * self.__Wh_to_Ws
        return capacity_dif / self.__timestep

    def update(self, energy_management_state: EnergyManagementState) -> None:
        pass

    def clear(self) -> None:
        pass

    def close(self) -> None:
        self.__soc_profile.close()



from abc import ABC, abstractmethod

from simses.commons.state.system import SystemState


class PowerDistributor(ABC):

    """
    PowerDistributor incorporates a logic on how to distribute power between several systems. The logic is based on the
    system state of each system.
    """

    def __init__(self):
        super().__init__()

    @abstractmethod
    def set(self, time: float, states: [SystemState], power: float) -> None:
        """
        Setting information from all system states necessary for the PowerDistributor

        Parameters
        ----------
        time :
            current simulation time as epoch time
        states :
            list of current system states
        power :
            overall power to be distributed to all systems
        """
        pass

    @abstractmethod
    def get_power_for(self, power: float, state: SystemState) -> float:
        """
        Calculates the power share of the overall to be distributed power to a specific system specified with system state

        Parameters
        ----------
        power :
            overall power to be distributed to all systems
        state :
            system state of system to calculate a specific power share of power

        Returns
        -------
        float:
            power share of specified system with corresponding system state
        """
        pass

    def close(self):
        pass



from simses.analysis.data.fuel_cell import FuelCellData
from simses.analysis.evaluation.plotting.axis import Axis
from simses.analysis.evaluation.plotting.plotly_plotting import PlotlyPlotting
from simses.analysis.evaluation.plotting.plotter import Plotting
from simses.analysis.evaluation.technical.technical_evaluation import TechnicalEvaluation
from simses.commons.config.analysis.general import GeneralAnalysisConfig
from simses.commons.state.technology.fuel_cell import FuelCellState


class FuelCellTechnicalEvaluation(TechnicalEvaluation):

    title = 'Fuel cell results'

    def __init__(self, data: FuelCellData, config: GeneralAnalysisConfig, path: str):
        super().__init__(data, config)
        title_extension: str = ' for system ' + self.get_data().id
        self.title += title_extension
        self.__result_path = path

    def evaluate(self) -> None:
        # super().evaluate()
        # data: FuelCellData = self.get_data()
        # self.append_result(EvaluationResult(Description.Technical.H2_PRODUCTION_EFFICIENCY_LHV, Unit.PERCENTAGE, self.h2_production_efficiency_lhv()))
        self.print_results()

    def plot(self) -> None:
        self.current_plotting()
        self.current_dens_plotting()
        self.pressures_plotting()
        self.temperature_plotting()
        self.power_plotting()
        self.hydrogen_consumption_plotting()

    def current_plotting(self):
        data: FuelCellData = self.get_data()
        plot: Plotting = PlotlyPlotting(title=FuelCellState.CURRENT, path=self.__result_path)
        xaxis: Axis = Axis(data=Plotting.format_time(data.time), label=FuelCellState.TIME)
        yaxis: [Axis] = [Axis(data=data.current, label=FuelCellState.CURRENT)]
        plot.lines(xaxis, yaxis)
        self.extend_figures(plot.get_figures())

    def current_dens_plotting(self):
        data: FuelCellData = self.get_data()
        plot: Plotting = PlotlyPlotting(title=FuelCellState.CURRENT_DENSITY, path=self.__result_path)
        xaxis: Axis = Axis(data=Plotting.format_time(data.time), label=FuelCellState.TIME)
        yaxis: [Axis] = [Axis(data=data.current_density, label=FuelCellState.CURRENT_DENSITY)]
        plot.lines(xaxis, yaxis)
        self.extend_figures(plot.get_figures())

    def pressures_plotting(self):
        data: FuelCellData = self.get_data()
        plot: Plotting = PlotlyPlotting(title='Pressures Cathode', path=self.__result_path)
        xaxis: Axis = Axis(data=Plotting.format_time(data.time), label=FuelCellState.TIME)
        yaxis: [Axis] = list()
        yaxis.append(Axis(data=data.pressure_cathode, label=FuelCellState.PRESSURE_CATHODE,
                          color=PlotlyPlotting.Color.CATHODE_PINK, linestyle=PlotlyPlotting.Linestyle.SOLID))
        yaxis.append(Axis(data=data.pressure_anode, label=FuelCellState.PRESSURE_ANODE,
                          color=PlotlyPlotting.Color.ANODE_GREEN, linestyle=PlotlyPlotting.Linestyle.SOLID))
        plot.lines(xaxis, yaxis)
        self.extend_figures(plot.get_figures())

    def temperature_plotting(self):
        data: FuelCellData = self.get_data()
        plot: Plotting = PlotlyPlotting(title=FuelCellState.TEMPERATURE, path=self.__result_path)
        xaxis: Axis = Axis(data=Plotting.format_time(data.time), label=FuelCellState.TIME)
        yaxis: [Axis] = list()
        yaxis.append(Axis(data=data.temperature, label=FuelCellState.TEMPERATURE, color=PlotlyPlotting.Color.TEMPERATURE_RED))
        plot.lines(xaxis, yaxis, [1])
        self.extend_figures(plot.get_figures())

    def power_plotting(self):
        data: FuelCellData = self.get_data()
        plot: Plotting = PlotlyPlotting(title=FuelCellState.POWER, path=self.__result_path)
        xaxis: Axis = Axis(data=Plotting.format_time(data.time), label=FuelCellState.TIME)
        yaxis: [Axis] = [Axis(data=data.power, label=FuelCellState.POWER)]
        plot.lines(xaxis, yaxis)
        self.extend_figures(plot.get_figures())

    def hydrogen_consumption_plotting(self):
        data: FuelCellData = self.get_data()
        plot: Plotting = PlotlyPlotting(title=FuelCellState.HYDROGEN_USE, path=self.__result_path)
        xaxis: Axis = Axis(data=Plotting.format_time(data.time), label=FuelCellState.TIME)
        yaxis: [Axis] = [Axis(data=data.hydrogen_use, label=FuelCellState.HYDROGEN_USE)]
        plot.lines(xaxis, yaxis)
        self.extend_figures(plot.get_figures())




import pytest

from simses.commons.state.system import SystemState
from simses.logic.power_distribution.soc import SocBasedPowerDistributor


@pytest.fixture()
def uut():
    return SocBasedPowerDistributor()


def create_state(soc: float) -> SystemState:
    state: SystemState = SystemState(0, 0)
    state.soc = soc
    return state


@pytest.mark.parametrize('power, soc_low, soc_high',
                         [
                             (0, 0, 0),
                             (0, 0, 1),
                             (0, 1, 1),
                             (10, 0, 0),
                             (10, 0.3, 0.7),
                             (10, 0, 1),
                             (10, 1, 1),
                             (10, 3, 5)
                         ]
                         )
def test_power(soc_high, soc_low, uut: SocBasedPowerDistributor, power):
    state_high = create_state(soc_high)
    state_low = create_state(soc_low)
    states: list = list()
    states.append(state_high)
    states.append(state_low)
    uut.set(0.0, states, 0)
    assert uut.get_power_for(power, state_low) >= uut.get_power_for(power, state_high)



from simses.commons.log import Logger
from simses.commons.config.simulation.redox_flow import RedoxFlowConfig
from simses.commons.state.technology.redox_flow import RedoxFlowState
from simses.system.auxiliary.pump.abstract_pump import Pump
from simses.technology.redox_flow.pump_algorithm.abstract_pump_algorithm import PumpAlgorithm
from simses.technology.redox_flow.stack.electrolyte.abstract_electrolyte import ElectrolyteSystem
from simses.technology.redox_flow.stack.abstract_stack import StackModule


class StoichFlowRate(PumpAlgorithm):
    """Pump algorithm with constant stoichimetric factor. The flow rate is calculated using the stoichimetric factor
    and the maximal still usable state-of-charge (SOC) difference for charging or discharging."""

    __STOICHIOMETRIC_FACTOR = 6

    def __init__(self, stack_module: StackModule, pump: Pump, electrolyte_system: ElectrolyteSystem,
                 redox_flow_config: RedoxFlowConfig):
        super().__init__(pump)
        self.__log: Logger = Logger(type(self).__name__)
        self.__stack_module: StackModule = stack_module
        self.__electrolyte_system: ElectrolyteSystem = electrolyte_system
        self.__redox_flow_config: RedoxFlowConfig = redox_flow_config
        self.__current_max_low_soc = self.__stack_module.get_max_current_low_soc()
        self.__current_max_high_soc = self.__stack_module.get_max_current_high_soc()
        self.__hydraulic_resistance = self.__stack_module.get_hydraulic_resistance()
        self.__stoichiometry = self.__STOICHIOMETRIC_FACTOR
        self.__FARADAY = self.__electrolyte_system.FARADAY

    def get_pressure_drop_catholyte(self, redox_flow_state: RedoxFlowState) -> float:
        pressure_drop = (self.get_flow_rate_catholyte(redox_flow_state) * self.__hydraulic_resistance *
                         self.__electrolyte_system.get_viscosity_catholyte(redox_flow_state))
        self.__log.debug('Pressure drop for catholyte: ' + str(pressure_drop / 10 ** 5) + ' bar')
        return pressure_drop

    def get_pressure_drop_anolyte(self, redox_flow_state: RedoxFlowState) -> float:
        pressure_drop = (self.get_flow_rate_anolyte(redox_flow_state) * self.__hydraulic_resistance *
                         self.__electrolyte_system.get_viscosity_anolyte(redox_flow_state))
        self.__log.debug('Pressure drop for anolyte: ' + str(pressure_drop / 10 ** 5) + ' bar')
        return pressure_drop

    def get_flow_rate_anolyte(self, redox_flow_state: RedoxFlowState) -> float:
        return self.get_flow_rate(redox_flow_state)

    def get_flow_rate_catholyte(self, redox_flow_state: RedoxFlowState) -> float:
        return self.get_flow_rate(redox_flow_state)

    def get_flow_rate(self, redox_flow_state: RedoxFlowState) -> float:
        """
        Calculates the flow rate based on an stoichiometric factor.

        Parameters
        ----------
        redox_flow_state : RedoxFlowState
            Current state of redox_flow.

        Returns
        -------
        float :
            Flow rate in m^3/s.
        """
        if redox_flow_state.is_charge:
            delta_soc = 1 - self.get_soc_begin()
        else:
            delta_soc = self.get_soc_begin()
        if delta_soc < 0.001:
            delta_soc = 0.001
        flow_rate = (self.__stoichiometry * abs(redox_flow_state.current) * self.__stack_module.get_cell_per_stack() *
                     self.__stack_module.get_serial_scale() / (self.__FARADAY *
                     self.__electrolyte_system.get_vanadium_concentration() * delta_soc))
        return flow_rate

    def get_flow_rate_max(self) -> float:
        delta_soc_min_low = self.__redox_flow_config.min_soc
        if delta_soc_min_low == 0:
            delta_soc_min_low = 0.001
        flow_rate_max_low_soc = (self.__stoichiometry * self.__current_max_low_soc *
                                 self.__stack_module.get_cell_per_stack() *
                                 self.__stack_module.get_serial_scale() / (self.__FARADAY *
                                 self.__electrolyte_system.get_vanadium_concentration() * delta_soc_min_low))
        delta_soc_min_high = (1 - self.__redox_flow_config.max_soc)
        if delta_soc_min_high == 0:
            delta_soc_min_high = 0.001
        flow_rate_max_high_soc = (self.__stoichiometry * self.__current_max_high_soc *
                                  self.__stack_module.get_cell_per_stack() *
                                  self.__stack_module.get_serial_scale() / (self.__FARADAY *
                                  self.__electrolyte_system.get_vanadium_concentration() * delta_soc_min_high))
        flow_rate_max = max(flow_rate_max_low_soc, flow_rate_max_high_soc)
        self.__log.debug('Max flow rate: ' + str(flow_rate_max))
        return flow_rate_max

    def get_flow_rate_min(self) -> float:
        flow_rate_min = 0.0
        return flow_rate_min

    def close(self):
        super().close()
        self.__log.close()
        self.__stack_module.close()
        self.__electrolyte_system.close()



from abc import abstractmethod, ABC

from simses.commons.constants import Hydrogen
from simses.commons.state.technology.electrolyzer import ElectrolyzerState


class PressureModel(ABC):
    """
    PressureModelEl calculates the pressure development at the cathode of the electrolyzer in dependency of the
    hydrogenproductionrate, the hydrogenoutflow and the temperature
    """

    def __init__(self):
        super().__init__()

    def update(self, time: float, state: ElectrolyzerState):
        self.calculate(time, state)
        state.pressure_cathode = self.get_pressure_cathode()  # bar
        state.pressure_anode = self.get_pressure_anode()  # bar   anode pressure is not varied -> stays constant
        state.hydrogen_outflow = self.get_h2_outflow()
        state.total_hydrogen_production += state.hydrogen_outflow * (time - state.time) * 2 * Hydrogen.MOLAR_MASS_HYDROGEN
        state.oxygen_outflow = self.get_o2_outflow()
        state.water_outflow_cathode = self.get_h2o_outflow_cathode()
        state.water_outflow_anode = self.get_h2o_outflow_anode()

    @abstractmethod
    def calculate(self, time, state: ElectrolyzerState) -> None:
        pass

    @abstractmethod
    def get_pressure_cathode(self) -> float:
        pass

    @abstractmethod
    def get_pressure_anode(self) -> float:
        pass

    @abstractmethod
    def get_h2_outflow(self) -> float:
        pass

    @abstractmethod
    def get_o2_outflow(self) -> float:
        pass

    @abstractmethod
    def get_h2o_outflow_cathode(self) -> float:
        pass

    @abstractmethod
    def get_h2o_outflow_anode(self) -> float:
        pass

    def close(self) -> None:
        """ Closing all resources in pressure model """
        pass


from random import Random

from simses.commons.profile.power.power_profile import PowerProfile


class RandomPowerProfile(PowerProfile):
    """
    RandomPowerProfile is a specific implementation of PowerProfile. It delivers a random power value for each timestep.
    Power series are reproducable - the used Random class uses the a specific seed for calculating random numbers.
    """

    def __init__(self, start_time: float, max_power: float = 1500.0, power_offset: float = 0.0,
                 scaling_factor: float = 1.0, time_increment: float = 1.0):
        """
        Constructor for RandomPowerProfile

        Parameters
        ----------
        max_power :
            cut off power in W for positive and negative values, default: 1500.0
        power_offset :
            power offset in W around which the random values are floating, default: 0.0
        scaling_factor :
            linear scaling of values, default: 1
        """
        super().__init__()
        self.__power: float = 0.0  # W
        self.__max_power: float = max_power  # W
        self.__d_power: float = max_power / 10.0  # W
        self.__random: Random = Random(93823341)
        self.__scaling_factor: float = scaling_factor
        self.__power_offset: float = power_offset
        self.__time: float = start_time
        self.__time_increment: float = time_increment

    def next(self, time: float) -> float:
        average: [float] = list()
        while self.__time <= time:
            dpower = self.__random.uniform(-self.__d_power, self.__d_power)
            self.__power = max(-self.__max_power, min(self.__max_power, self.__power + dpower))
            average.append(self.__power)
            self.__time += self.__time_increment
        power = sum(average) / len(average) if average else self.__power
        return power * self.__scaling_factor + self.__power_offset

    def close(self) -> None:
        pass



from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.profile.file import FileProfile
from simses.system.thermal.ambient.ambient_thermal_model import AmbientThermalModel
from simses.commons.config.simulation.profile import ProfileConfig


class UserBatteryTemperatureProfile(AmbientThermalModel):

    """
    UserBatteryTemperatureProfile reads a user-defined battery temperature profile for the specified battery.
    A temperature value for each timestep is read and updated in the state of the storage technology
    """

    def __init__(self, profile_config: ProfileConfig, general_config: GeneralSimulationConfig):
        super().__init__()
        self.__start_time = general_config.start
        self.__file = FileProfile(general_config, profile_config.battery_temperature_profile_file)
        self.__profile_config: ProfileConfig = profile_config
        self.__general_config: GeneralSimulationConfig = general_config

    def get_temperature(self, time) -> float:
        temp_c = self.__file.next(time)  # in °C
        temp_k = temp_c + 273.15  # in K
        return temp_k

    def get_initial_temperature(self) -> float:
        return self.get_temperature(self.__start_time)  # in K

    def create_instance(self) -> AmbientThermalModel:
        """
        reinstantiates the AmbientThermalModel
        :return: AmbientThermalModel
        """
        return UserBatteryTemperatureProfile(self.__profile_config, self.__general_config)

    def close(self):
        self.__file.close()


from datetime import datetime

from simses.system.auxiliary.auxiliary import Auxiliary
from simses.system.dc_coupling.load.dc_load import DcLoad


class DCRadioUPSLoad(DcLoad):

    """
    DcBusChargingFixed acts as fixed load profile for bus connecting to the storage system at the same time each day.
    """

    def __init__(self, charging_power: float):
        super().__init__()
        self.__Load_by_Night: float = charging_power  # W
        self.__Load: float = 800.0

    def get_power(self) -> float:
        return self.__Load

    def calculate_power(self, time: float) -> None:
        tstmp = datetime.utcfromtimestamp(time)
        hour = tstmp.hour + tstmp.minute / 60. + tstmp.second / 3600.
        if 20 <= hour <= 24:
            self.__Load = self.__Load_by_Night
        elif 0 <= hour <= 8:
            self.__Load = self.__Load_by_Night
        else:
            self.__Load = 800.0


    def get_auxiliaries(self) -> [Auxiliary]:
        return list()

    def close(self) -> None:
        pass


from math import ceil, floor
import matplotlib.pyplot as plt
from simses.analysis.evaluation.plotting.axis import Axis
from simses.analysis.evaluation.plotting.plotter import Plotting


class MatplotPlotting(Plotting):

    __nbins=100

    class Linestyle:
        DOTTED = 'dotted'
        SOLID = 'solid'
        DASHED = 'dashed'
        DASH_DOT = 'dashdot'

    def __init__(self, title: str=" ", path: str=" "):
        super().__init__()
        self.__title = title
        self.__path = path

    def lines(self, xaxis: Axis, yaxis: [Axis], secondary=[]):
        plt.title(self.__title)
        # plt.ylabel(self.__ylabel)
        plt.xlabel(xaxis.label)
        for y in yaxis:
            plt.plot(xaxis.data, y.data, color=y.color, linestyle=y.linestyle, label=y.label)
        plt.legend()
        plt.grid()
        self.show()

    def show(self):
        plt.savefig(self.__path + self.alphanumerize(self.__title)+'.pdf')
        plt.show()

    def histogram(self, xaxis: Axis, yaxis: [Axis]):
        cols = ceil(len(yaxis) / 2)
        rows = ceil(len(yaxis) / 2)
        fig, axes = plt.subplots(nrows=rows, ncols=cols, figsize=(16, 9), squeeze=False)
        fig.suptitle(self.__title)
        for x in range(len(yaxis)):
            ydata = yaxis[x]
            row = floor(x / rows)
            col = (x % cols)
            axes[row, col].hist(ydata.data, bins=self.__nbins, density=True, color = ydata.color)
            axes[row, col].set(xlabel=ydata.label)
        self.show()

    def bar(self, yaxis: [Axis], bars: int):
        variables = [yaxis[x:x + bars] for x in range(0, len(yaxis), bars)]
        titles = [yaxis[x].label.split(" ", 1)[1] for x in range(0, len(yaxis), 2)]
        cols = ceil(len(yaxis) / bars)
        rows = ceil(len(yaxis) / (bars + cols))
        fig, axes = plt.subplots(nrows=rows, ncols=cols, figsize=(16, 9), squeeze=False)
        fig.suptitle(self.__title)
        for i in range(len(variables)):
            variable = variables[i]
            labels = []
            values = []
            colors = []
            for j in range(len(variable)):
                labels.append(variable[j].label.split()[0])
                values.append(variable[j].data.sum())
                colors.append(variable[j].color)
            axes[ceil((i + j) / 4)-1, ceil((i + j) / 2)-1].bar(x=labels,height=values,color=colors)
            axes[ceil((i + j) / 4)-1, ceil((i + j) / 2)-1].set_title(titles[i])
        self.show()

    def subplots(self, xaxis: Axis, yaxis: [Axis]):
        cols = ceil(len(yaxis) / 2)
        rows = ceil(len(yaxis) / 2)
        fig, axes = plt.subplots(nrows=rows, ncols=cols, figsize=(16, 9), squeeze=False)
        fig.suptitle(self.__title)
        for x in range(len(yaxis)):
            ydata = yaxis[x]
            row = floor(x / rows)
            col = (x % cols)
            axes[row, col].plot(xaxis.data, ydata.data)
            axes[row, col].set(xlabel=xaxis.label, ylabel='', title=ydata.label)
            axes[row, col].grid()
        self.show()

    def sankey_diagram(self, node_links: dict):
        pass

    def sunburst_diagram(self, categories: dict):
        pass


from simses.commons.cycle_detection.cycle_detector import CycleDetector
from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.technology.lithium_ion.cell.type import CellType
from simses.technology.lithium_ion.degradation.cyclic.cyclic_degradation import CyclicDegradationModel


class NMCGenericCyclicDegradationModel(CyclicDegradationModel):

    def __init__(self, cell_type: CellType, cycle_detector: CycleDetector):
        super().__init__(cell_type, cycle_detector)
        self.__capacity_loss = 0
        self.__initial_capacity = self._cell.get_nominal_capacity()

        # parameters old
        # self.__z = 0.705799526688744
        # self.__a = -0.0305245206114891
        # self.__b = 27.6760345173264
        # self.__c = -8353.74758895471
        # self.__d = 839529.464749593
        # self.__e = -1.03958438129589
        # self.__f = 2.95820190976433
        # self.__g = -2.06426540159782
        # self.__h = 0.436349803729391
        # self.__i = -0.183172940789106
        # self.__j = 3.77584036178456
        # self.__k = -2.02237532152262
        # self.__l = 0.256120170615456
        # self.__m = 0.000819809707717746
        # self.__n = 0.0302888311700094

        # parameters new
        self.__z = 0.671627179869084
        self.__a = 2.9957641942448E-07
        self.__b = -0.0000720878146452049
        self.__c = -0.0385306420008836
        self.__d = 9.99999995479674
        self.__e = -3.02499312835016
        self.__f = 8.81907826557735
        self.__g = -6.19696246135569
        self.__h = 1.31642579445754
        self.__i = -1.24637880290245
        self.__j = 9.95196984230208
        self.__k = -5.69035010249573
        self.__l = 0.870681264163946
        self.__m = 0.270719035641741
        self.__n = 9.99676141875715

    def fec_dependency(self, fec):
        # degradation depends on fec as a power law function
        return fec**self.__z

    def temperature_dependency(self, temp):
        # degradation depends on temperature as a cubic function
        return self.__a * temp**3 + self.__b * temp**2 + self.__c * temp + self.__d

    def mean_soc_dependency(self, msoc):
        # degradation depends on msoc as a cubic function
        return self.__e * msoc**3 + self.__f * msoc**2 + self.__g * msoc + self.__h

    def doc_dependency(self, doc):
        # degradation depends on doc as a cubic function
        return self.__i * doc ** 3 + self.__j * doc ** 2 + self.__k * doc + self.__l

    def crate_dependency(self, crate):
        # degradation depends on crate as a linear function
        return self.__m * crate + self.__n

    def calculate_degradation(self, battery_state: LithiumIonState) -> None:
        c_rate: float = self._cycle_detector.get_crate() * 3600
        depth_of_discharge: float = self._cycle_detector.get_depth_of_cycle()
        full_equivalent_cycle: float = self._cycle_detector.get_full_equivalent_cycle()
        mean_soc: float = self._cycle_detector.get_mean_soc()
        temperature: float = battery_state.temperature
        delta_fec: float = self._cycle_detector.get_delta_full_equivalent_cycle()

        # calculate degradation in percent (divide by 100 to get p.u.) with function 29 from MA Sandner
        capacity_loss_after = self.fec_dependency(full_equivalent_cycle) * \
                              self.temperature_dependency(temperature) * \
                              self.mean_soc_dependency(mean_soc) * \
                              self.doc_dependency(depth_of_discharge) * \
                              self.crate_dependency(c_rate)
        capacity_loss_before = self.fec_dependency(full_equivalent_cycle - delta_fec) * \
                               self.temperature_dependency(temperature) * \
                               self.mean_soc_dependency(mean_soc) * \
                               self.doc_dependency(depth_of_discharge) * \
                               self.crate_dependency(c_rate)

        self.__capacity_loss = (capacity_loss_after - capacity_loss_before) / 100.0 * self.__initial_capacity

    def calculate_resistance_increase(self, battery_state: LithiumIonState) -> None:
        # no resistance increase calculated for generic degradation model
        pass

    def get_degradation(self) -> float:
        degradation = self.__capacity_loss
        self.__capacity_loss = 0
        return degradation

    def get_resistance_increase(self) -> float:
        # no resistance increase calculated for generic degradation model
        return 0

    def reset(self, lithium_ion_state: LithiumIonState) -> None:
        self.__capacity_loss = 0

    def close(self) -> None:
        pass



import os
from configparser import ConfigParser
from datetime import datetime, timezone
import pytest
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.config.simulation.profile import ProfileConfig
from simses.commons.utils.utilities import remove_file
from simses.system.thermal.ambient.location_temperature import LocationAmbientTemperature
from simses.system.thermal.ambient.user_battery_temperature import UserBatteryTemperatureProfile

DELIMITER: str = ','
HEADER: str = '# Unit: C'
START: int = 0
END: int = 60
STEP: int = 1


def create_file(value, string:str = '1') -> str:
    filename = string + 'mockup_ambient_temperature_profile.csv'
    path: str = os.path.join(os.path.dirname(__file__), filename)
    with open(path, mode='w') as file:
        file.write(HEADER + '\n')
        # add inputs
        tstmp = START
        while tstmp <= END:
            file.write(str(tstmp) + DELIMITER + str(value) + '\n')
            tstmp += STEP
    return path


def create_general_config() -> GeneralSimulationConfig:
    conf: ConfigParser = ConfigParser()
    conf.add_section(GeneralSimulationConfig.SECTION)
    date = datetime.fromtimestamp(START, timezone.utc)
    conf.set(GeneralSimulationConfig.SECTION, GeneralSimulationConfig.START, date.strftime("%Y-%m-%d %H:%M:%S"))
    date = datetime.fromtimestamp(END, timezone.utc)
    conf.set(GeneralSimulationConfig.SECTION, GeneralSimulationConfig.END, date.strftime("%Y-%m-%d %H:%M:%S"))
    conf.set(GeneralSimulationConfig.SECTION, GeneralSimulationConfig.TIME_STEP, str(STEP))
    return GeneralSimulationConfig(conf)


def create_profile_config(filename: str) -> ProfileConfig:
    conf: ConfigParser = ConfigParser()
    conf.add_section(ProfileConfig.SECTION)
    conf.set(ProfileConfig.SECTION, ProfileConfig.THERMAL_PROFILE_DIR, os.path.dirname(filename))
    conf.set(ProfileConfig.SECTION, ProfileConfig.AMBIENT_TEMPERATURE_PROFILE, os.path.basename(filename))
    return ProfileConfig(conf)


def create_profile_config_2(filename: str) -> ProfileConfig:
    conf: ConfigParser = ConfigParser()
    conf.add_section(ProfileConfig.SECTION)
    conf.set(ProfileConfig.SECTION, ProfileConfig.THERMAL_PROFILE_DIR, os.path.dirname(filename))
    conf.set(ProfileConfig.SECTION, ProfileConfig.BATTERY_TEMPERATURE_PROFILE, os.path.basename(filename))
    return ProfileConfig(conf)


@pytest.fixture(scope='function')
def uut(ambient_temperature):
    filename = create_file(ambient_temperature)
    uut = LocationAmbientTemperature(create_profile_config(filename), create_general_config())
    yield uut
    uut.close()
    remove_file(filename)


@pytest.fixture(scope='function')
def uut2(ambient_temperature):
    filename = create_file(ambient_temperature, string='2')
    uut2 = UserBatteryTemperatureProfile(create_profile_config_2(filename), create_general_config())
    yield uut2
    uut2.close()
    remove_file(filename)


@pytest.mark.parametrize('ambient_temperature, time_factor, result',
                         [
                             (0, 1, 0),
                             (0, 1, 0),
                             (1, 1, 1),
                             (2, 1, 2),
                             (6, 2, 6),
                             (1, 5, 1),
                             (-1, 5, -1),
                             (-15, 4, -15),
                             (-25,10,-25)
                         ]
                         )
def test_get_temperature(time_factor, result, uut, uut2):
    tstmp = START
    while tstmp < END:
        assert abs(uut.get_temperature(tstmp) - (result + 273.15)) <= 1e-10
        assert abs(uut2.get_temperature(tstmp) - (result + 273.15)) <= 1e-10
        tstmp += STEP * time_factor


import gzip
import os
import sys
from datetime import datetime
from os.path import basename
from pytz import timezone
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.error import EndOfFileError
from simses.commons.log import Logger
from simses.commons.timeseries.average.average import Average
from simses.commons.timeseries.average.mean_average import MeanAverage
from simses.commons.timeseries.interpolation.interpolation import Interpolation
from simses.commons.timeseries.interpolation.linear_interpolation import LinearInterpolation
from simses.commons.timeseries.timevalue import TimeValue
from simses.commons.utils.utilities import download_file


class FileProfile:
    """
    FileProfile is able to read time series from a file. It supports several unit and date formats, takes timezones into
    consideration (default: Berlin) and interpolates respectively averages values of the given time series. If no time series
    is given, a time generator generates the timestep according to the given sampling rate or simulation timestep.

    A header could inherit the given timezone, value unit, time unit and sampling rate:\n
    # Unit: W\n
    # Timezone: Berlin\n
    # Time: s\n
    # Sampling in s: 3600s\n
    """

    class Header:
        TIMEZONE: str = 'Timezone'
        UNIT: str = 'Unit'
        TIME: str = 'Time'
        SAMPLING: str = 'Sampling in s'
        LATITUDE: str = 'Latitude'
        LONGITUDE: str = 'Longitude'

    __UNITS: dict = {'W': 1, 'kW': 1e3, 'MW': 1e6, 'GW': 1e9, 's': 1, 'ms': 0.001, 'EUR': 1, 'TEUR': 1e3, 'Hz': 1, '%': 0.01, 'W/m2': 1, 'C': 1}
    """List of recognized unit formats of FileProfile"""

    __TIME_IDX: int = 0

    __EPOCH_FORMAT: str = 'epoch'
    __DATE_FORMATS: [str] = ['%Y-%m-%d %H:%M:%S', '%Y-%m-%d %H:%M', '%d.%m.%Y %H:%M:%S', '%d.%m.%Y %H:%M',
                             __EPOCH_FORMAT]
    """List of supported date formats"""

    __UTC: timezone = timezone('UTC')
    __BERLIN: timezone = timezone('Europe/Berlin')

    __EXT_GZIP: str = '.gz'
    __EXT_CSV: str = '.csv'
    __SUPPORTED_EXTENSIONS: [str] = list()
    __SUPPORTED_EXTENSIONS.append(__EXT_CSV + __EXT_GZIP)
    __SUPPORTED_EXTENSIONS.append(__EXT_GZIP)
    __SUPPORTED_EXTENSIONS.append(__EXT_CSV)

    __ERROR = None

    __URL: str = 'https://syncandshare.lrz.de/dl/fiACpACzuTyoENQE2qwqrG/profiles/'
    """URL to folder of profiles to download"""

    def __init__(self, config: GeneralSimulationConfig, filename: str, delimiter: str = ',', scaling_factor: float = 1,
                 value_index: int = 1, interpolation: Interpolation = LinearInterpolation(),
                 average: Average = MeanAverage()):
        """
        Constructor of FileProfile

        Parameters
        ----------
        config :
            simulation configuration
        filename :
            path to file
        delimiter :
            delimiter of values in file, default: ,
        scaling_factor :
            linear scaling of values, default: 1
        value_index :
            column index for values, default: 1
        interpolation:
            provides an Interpolation object, default: LinearInterpolation
        average:
            provides an Average object, default: MeanAverage
        """
        self.__log: Logger = Logger(type(self).__name__)
        self.__log.info('Preparing ' + filename)
        self.__filename: str = filename
        self.__delimiter: str = delimiter
        self.__scaling_factor: float = scaling_factor
        self.__VALUE_IDX: int = value_index
        self.__timestep: float = config.timestep
        self.__start: float = config.start
        self.__end: float = config.end
        self.__time_offset: int = 0
        self.__file = None
        self.__date_format = None
        self.__last_data: [TimeValue] = list()
        self.__last_time: float = self.__start
        self.__interpolation: Interpolation = interpolation
        self.__average: Average = average
        header: dict = self.get_header_from(self.__filename)
        self.__unit_factor: float = self.__get_value_unit_from(header)
        self.__time_factor: float = self.__get_time_unit_from(header)
        self.__timezone: timezone = self.__get_timezone_from(header)
        self.__sampling_time: float = self.__get_sampling_time_from(header)
        self.__time_generator = self.__get_sampling_time_generator()
        self.__latitude: float = self.__get_latitude_from(header)
        self.__longitude: float = self.__get_longitude_from(header)
        self.initialize_file()

    def next(self, time: float) -> float:
        """
        Retrieves the next value for given time from file time series

        Parameters
        ----------
        time :
            time as epoch timestamp

        Returns
        -------
        float :
            (interpolated/averaged) value for given time
        """
        try:
            data: [TimeValue] = self.__get_data_until(time)
            values: [TimeValue] = self.__filter_current_values(time, data)
            if not values and len(data) < 2:
                value: float = self.__average.average(data)
            elif not values:
                value: float = self.__interpolation.interpolate(time, data[-1], data[-2])
            else:
                value: float = self.__average.average(values)
            self.__last_data = self.__filter_last_values(data)
            self.__last_time = time
            return value * self.__scaling_factor
        except EndOfFileError as err:
            self.__log.warn(str(err))
            return 0.0

    def initialize_file(self) -> None:
        """
        Allows to re-initialize the profile to start reading values from the beginning again
        """
        if self.__file is not None:
            if not self.__file.closed:
                self.__file.close()
        self.__file = self.open_file(self.__filename)
        self.__last_data = self.__get_initial_data()

    @classmethod
    def open_file(cls, filename: str):
        """
        Opens file with filename depending on the file type. Currently .gz and .csv are supported.
        The file type extension is added if necessary.

        Parameters
        ----------
        filename :
            path to file

        Returns
        -------
            file object (needs to be closed manually)
        """
        filename: str = cls.__add_extension_to(filename)
        try:
            if filename.endswith(cls.__EXT_GZIP):
                return gzip.open(filename, 'rt')
            elif filename.endswith(cls.__EXT_CSV):
                return open(filename, 'r', newline='')
            else:
                raise FileNotFoundError('Format of ' + filename + ' is currently not supported. Please check your '
                                            'file type. Following file types are supported: ' + str(cls.__SUPPORTED_EXTENSIONS))
        except FileNotFoundError as err:
            if cls.__URL is cls.__ERROR:
                raise err
            print(filename + ' is not available. Searching for file on server...')
            if cls.__profile_downloaded(filename):
                return cls.open_file(filename)
            else:
                raise FileNotFoundError(basename(filename) + ' is currently not available for download.', err)

    @classmethod
    def __profile_downloaded(cls, filename: str) -> bool:
        found: bool = False
        for file in cls.__possible_file_names(filename):
            try:
                download_file(cls.__URL + basename(file), file)
                found = True
                break
            except FileNotFoundError:
                print(basename(file) + ' could not be found on server')
        return found

    @classmethod
    def __possible_file_names(cls, filename: str) -> [str]:
        possible_file_names: [str] = list()
        if cls.__has_extension(filename):
            possible_file_names.append(filename)
        else:
            for extension in cls.__SUPPORTED_EXTENSIONS:
                possible_file_names.append(filename + extension)
        return possible_file_names

    @classmethod
    def __has_extension(cls, filename: str) -> bool:
        has_extension: bool = False
        for extension in cls.__SUPPORTED_EXTENSIONS:
            has_extension |= filename.endswith(extension)
        return has_extension

    @classmethod
    def __add_extension_to(cls, filename: str) -> str:
        has_extension: bool = cls.__has_extension(filename)
        if not has_extension:
            for extension in cls.__SUPPORTED_EXTENSIONS:
                if os.path.isfile(filename + extension):
                    filename += extension
                    break
        return filename

    def __filter_current_values(self, time: float, data: [TimeValue]) -> [TimeValue]:
        values: [TimeValue] = list()
        data_iterator = iter(data)
        try:
            while True:
                value = next(data_iterator)
                if self.__last_time < value.time <= time:
                    values.append(value)
        except StopIteration:
            pass
        return values

    def __filter_last_values(self, data: [TimeValue]) -> [TimeValue]:
        return data[-2:]
        # values: [TimeValue] = list()
        # data_iterator = iter(data)
        # try:
        #     while True:
        #         value = next(data_iterator)
        #         if self.__last_time < value.time:
        #             values.append(value)
        # except StopIteration:
        #     pass
        # return values

    def __get_data_until(self, time: float) -> [TimeValue]:
        values: list = list()
        data = self.__last_data[-1]
        while data.time < time:
            data = self.__get_next_data()
            values.append(data)
        values.extend(self.__last_data)
        TimeValue.sort_by_time(values)
        return values

    def __get_next_data(self) -> TimeValue:
        while True:
            line: str = ''
            try:
                line = self.__file.readline()
                if line.startswith('#') or line.startswith('"') or line in ['\r\n', '\n']:# or self.__delimiter not in line:
                    continue
                if line == '':  # end of file_name
                    raise EndOfFileError('End of Profile ' + self.__filename + ' reached.')
                data = line.split(self.__delimiter)
                if len(data) < 2:
                    time: float = next(self.__time_generator) + self.__time_offset
                    value: float = float(data[self.__VALUE_IDX - 1]) * self.__unit_factor
                else:
                    time: float = self.__format_time(data[self.__TIME_IDX]) * self.__time_factor + self.__time_offset
                    value: float = float(data[self.__VALUE_IDX]) * self.__unit_factor
                return TimeValue(time, value)
            except ValueError:
                self.__log.error('No value found for ' + line)

    def __get_sampling_time_generator(self):
        return self.__get_time_generator(self.__sampling_time)

    def __get_timestep_time_generator(self):
        return self.__get_time_generator(self.__timestep)

    def __get_time_generator(self, time_increment: float):
        time: float = self.__start
        while True:
            yield time
            time += time_increment

    def __format_time(self, time: str) -> float:
        if self.__date_format is None:
            self.__date_format = self.__find_date_format_for(time)
            self.__log.info('Found format: ' + str(self.__date_format))
        if self.__date_format == self.__EPOCH_FORMAT:
            return float(time)
        else:
            return self.__extract_timestamp_from(time, self.__date_format)

    def __find_date_format_for(self, time: str) -> str:
        for date_format in self.__DATE_FORMATS:
            try:
                if date_format == self.__EPOCH_FORMAT:
                    float(time)
                    return date_format
                else:
                    self.__extract_timestamp_from(time, date_format)
                    return date_format
            except ValueError:
                pass
        raise Exception('Unknown date format for ' + time)

    def __extract_timestamp_from(self, time: str, date_format: str) -> float:
        date: datetime = datetime.strptime(time, date_format)
        date = self.__get_local_datetime_from(date=date)
        return date.timestamp()

    def __get_local_datetime_from(self, date: datetime = None, tstmp: float = None) -> datetime:
        if date is None:
            if tstmp is None:
                tstmp = datetime.now()
            date = datetime.fromtimestamp(tstmp)
        return self.__timezone.localize(date, is_dst=None)

    def __get_initial_data(self) -> [TimeValue]:
        timestamp: float = self.__start
        data = self.__get_next_data()
        self.__set_time_offset(data.time, timestamp)
        data.time += self.__time_offset
        while data.time < timestamp:
            data = self.__get_next_data()
        return [data]

    def __set_time_offset(self, file_tstmp: float, simulation_tstmp: float) -> None:
        #Set profile year to simulation year
        file_date = self.__get_local_datetime_from(tstmp=file_tstmp)
        simulation_date = self.__get_local_datetime_from(tstmp=simulation_tstmp)
        if file_date.year == simulation_date.year:
            return
        adapted_file_tstmp = file_date.replace(year=simulation_date.year).timestamp()
        self.__time_offset = adapted_file_tstmp - file_tstmp
        if not self.__time_offset == 0:
            self.__log.warn('Time offset is ' + str(self.__time_offset) + ' s. \n'
                            'File time: ' + str(self.__get_local_datetime_from(tstmp=file_tstmp)) + ', \n'
                            'Simulation time: ' + str(self.__get_local_datetime_from(tstmp=simulation_tstmp)))

    def __get_timezone_from(self, header: dict) -> timezone:
        try:
            tz: str = header[self.Header.TIMEZONE]
            return timezone(tz)
        except KeyError:
            self.__log.warn('No timezone for profile is given, UTC timzone is assumed.')
            return self.__UTC

    def __get_latitude_from(self, header: dict) -> float:
        try:
            lat: str = header[self.Header.LATITUDE]
            return float(lat)
        except KeyError:
            return self.__ERROR

    def __get_longitude_from(self, header: dict) -> float:
        try:
            lon: str = header[self.Header.LONGITUDE]
            return float(lon)
        except KeyError:
            return self.__ERROR

    @classmethod
    def __get_unit_from(cls, header: dict, identifier: str) -> float:
        try:
            unit: str = header[identifier]
            return cls.__UNITS[unit]
        except KeyError:
            # print('Unit is unknown. Valid types are ' + str(cls.__UNITS.keys()) + '.')
            return 1.0

    @classmethod
    def __get_value_unit_from(cls, header: dict) -> float:
        return cls.__get_unit_from(header, cls.Header.UNIT)

    @classmethod
    def __get_time_unit_from(cls, header: dict) -> float:
        return cls.__get_unit_from(header, cls.Header.TIME)

    def __get_sampling_time_from(self, header: dict) -> float:
        try:
            sampling_time: str = header[self.Header.SAMPLING]
            return float(sampling_time)
        except KeyError:
            return self.__timestep

    def profile_data_to_list(self, sign_factor=1) -> ([float], [float]):
        """
        Extracts the whole time series as a list and resets the pointer of the (internal) file afterwards

        Parameters
        ----------
        sign_factor :

        Returns
        -------
        list:
            profile values as a list

        """
        profile_data: [float] = list()
        time_data: [float] = list()
        time_generator = self.__get_timestep_time_generator()
        time: float = next(time_generator)
        while time <= self.__end:
            time_data.append(time)
            profile_data.append(self.next(time) * sign_factor)
            time = next(time_generator)
        self.initialize_file()
        return time_data, profile_data

    def get_latitude(self) -> float:
        """

        Returns
        -------
        float:
            Latitude value of profile location, raises an exception if not available
        """
        if self.__latitude is self.__ERROR:
            raise Exception('No latitude value given in profile. Please add the header for latitude.')
        return self.__latitude

    def get_longitude(self) -> float:
        """

        Returns
        -------
        float:
            Longitude value of profile location, raises an exception if not available
        """
        if self.__longitude is self.__ERROR:
            raise Exception('No longitude value given in profile. Please add the header for longitude.')
        return self.__longitude

    def get_timezone(self) -> timezone:
        """

        Returns
        -------
        float:
            timezone for profile location, raises an exception if not available
        """
        if self.__timezone is self.__ERROR:
            raise Exception('No timezone value given in profile. Please add the header for timezone.')
        return self.__timezone

    @classmethod
    def get_header_from(cls, filename: str) -> dict:
        """
        Extracts header from given file

        Attention: Only searches in the first ten lines for a header!

        Parameters
        ----------
        filename :

        Returns
        -------
        dict:
            header with key/value pairs

        """
        header: dict = dict()
        file = cls.open_file(filename)
        # with open(filename, 'r', newline='') as file:
        line = file.readline()
        line_count: int = 0
        while True:
            if '#' in line:
                try:
                    key_raw, entry_raw = line.split(sep=':', maxsplit=1)
                    key = key_raw.strip('# ')
                    entry = entry_raw.strip()
                    header[key] = entry
                except ValueError:
                    sys.stderr.write('WARNING: Could not interpret ' + line)
                    sys.stderr.flush()
            line = file.readline()
            if line_count > 20:
                break
            else:
                line_count += 1
        file.close()
        return header

    @classmethod
    def get_max_value_of(cls, filename: str, delimiter: str = ',', value_idx: int = 1, start: float = 0.0) -> float:
        """
        Searching for maximum value within given file

        Attention: This method could be very time consuming depending on the profile to read.

        Parameters
        ----------
        filename :
            filename of profile which is read in
        delimiter :
            delimiter for values in one line, default: ','
        value_idx :
            column number of line for value which should be considered, default: 1
        start :
            start value for search

        Returns
        -------
        float:
            maximum value of given profile
        """
        # ts_before = time.time()
        header: dict = cls.get_header_from(filename)
        unit_factor: float = cls.__get_value_unit_from(header)
        file = cls.open_file(filename)
        max_value: float = start
        for line in file:
            if line.startswith('#') or line.startswith('"') or line in ['\r\n', '\n']:
                continue
            data: [str] = line.split(delimiter)
            if len(data) < 2:
                value: float = float(data[0])
            else:
                value: float = float(data[value_idx])
            max_value = max(value * unit_factor, max_value)
        file.close()
        # print(max_value, unit_factor)
        # print(str(time.time() - ts_before))
        return max_value

    def close(self):
        self.__log.close()
        self.__file.close()



from simses.system.auxiliary.gas_drying.gas_dryer import GasDrying


class HydrogenGasDrying(GasDrying):

    def __init__(self):
        super().__init__()

        self.__spec_drying_energy = 139.62 * 10 ** 3  #  J/mol h2o  from: Life cycle assessment of hydrogen from proton exchange membrane water electrolysis in future energy systems
        self.__gas_drying_power = 0  #  W
        self.x_dry = 5 / 1000000  # 5 ppm

    def calculate_gas_drying_power(self, pressure_cathode, h2_outflow) -> None:
        x_h2o_after_condens = 0.07981 * pressure_cathode ** (-1.04)  # after condensation at 40°C after curve of: PEM-Elektrolyse-Systeme zur Anwendung in Power-to-Gas Anlagen
        n_h2o_drying = (x_h2o_after_condens - self.x_dry) * h2_outflow  # mol/s
        self.__gas_drying_power = self.__spec_drying_energy * n_h2o_drying  # W

    def get_gas_drying_power(self) -> float:
        return self.__gas_drying_power



from datetime import datetime

import numpy
import numpy_financial

from simses.analysis.data.system import SystemData
from simses.analysis.evaluation.abstract_evaluation import Evaluation
from simses.analysis.evaluation.economic.revenue_stream.abstract_revenue_stream import RevenueStream
from simses.analysis.evaluation.result import EvaluationResult, Description, Unit
from simses.commons.config.analysis.economic import EconomicAnalysisConfig
from simses.commons.config.analysis.general import GeneralAnalysisConfig
from simses.commons.config.simulation.system import StorageSystemConfig
from simses.commons.log import Logger
from simses.commons.utils.utilities import format_float


class EconomicEvaluation(Evaluation):
    """
    Performs an economic evaluation by iterating through the respective RevenueStreams. Calculates key KPIs such as
     internal rate of return (IRR), net present value (NPV), etc..
    """

    def __init__(self, system_data: SystemData, economic_analysis_config: EconomicAnalysisConfig,
                 revenue_streams: [RevenueStream], config: GeneralAnalysisConfig,
                 storage_system_config: StorageSystemConfig):
        super().__init__(system_data, config, config.economical_analysis)
        self.__log: Logger = Logger(type(self).__name__)
        self.__config: GeneralAnalysisConfig = config
        self.__revenue_streams = revenue_streams
        self.__discount_rate = economic_analysis_config.discount_rate
        if economic_analysis_config.use_specific_costs:
            energy_costs: float = economic_analysis_config.specific_investment_costs_energy
            power_costs: float = economic_analysis_config.specific_investment_costs_power
            energy: float = system_data.initial_capacity / system_data.initial_state_of_health
            # TODO Review: Remove interdependence with StorageSystemConfig
            power: float = sum([float(storage_system_ac[StorageSystemConfig.AC_SYSTEM_POWER])
                                for storage_system_ac in storage_system_config.storage_systems_ac]) * 1e-3
            self.__investment_costs = energy_costs * energy + power_costs * power
        else:
            self.__investment_costs = economic_analysis_config.investment_costs
        # TODO clarify if this is the right way for coping with investment costs
        for stream in self.__revenue_streams:
            stream.set_investment_cost(self.__investment_costs)

    def evaluate(self):
        if self.__revenue_streams:
            # Calculate cashflow
            future_cashflow: numpy.ndarray = numpy.array([])
            for revenue_stream in self.__revenue_streams:
                cashflow: numpy.ndarray = revenue_stream.get_cashflow()
                self.extend_results(revenue_stream.get_evaluation_results())
                # get the right array size based on first incoming cash flow
                if not future_cashflow.size:
                    future_cashflow.resize(cashflow.shape, refcheck=False)
                # add cashflow and correct potential size issues for leap years
                if len(future_cashflow) < len(cashflow):
                    self.__log.warn('Cashflow arrays have different lengths.')
                future_cashflow[:len(cashflow)] += cashflow
            # Compile list of results
            self.append_result(EvaluationResult(Description.Economical.INVESTMENT_COSTS, Unit.EURO,
                                                self.__investment_costs))
            self.append_result(EvaluationResult(Description.Economical.CASHFLOW, Unit.EURO, future_cashflow))
            self.append_result(EvaluationResult(Description.Economical.NET_PRESENT_VALUE, Unit.EURO,
                                                self.__net_present_value(future_cashflow)))
            self.append_result(EvaluationResult(Description.Economical.INTERNAL_RATE_OF_RETURN, Unit.PERCENTAGE,
                                                self.__internal_rate_of_return(future_cashflow) * 100))
            self.append_result(EvaluationResult(Description.Economical.PROFITABILITY_INDEX, Unit.NONE,
                                                self.__profitability_index(future_cashflow)))
            self.append_result(EvaluationResult(Description.Economical.RETURN_ON_INVEST, Unit.NONE,
                                                self.__return_on_investment(future_cashflow)))
            self.append_result(EvaluationResult(Description.Economical.LEVELIZED_COST_OF_STORAGE, Unit.EURO_PER_MWH,
                                                self.__levelized_cost_of_storage_energy()))
            self.print_results()

    def plot(self) -> None:
        pass

    def __internal_rate_of_return(self, cashflow: numpy.ndarray) -> float:
        cashflow = numpy.append(numpy.array([-self.__investment_costs]), cashflow)
        if numpy.any(numpy.isnan(cashflow)) or numpy.any(numpy.isinf(cashflow)):
            self.__log.error('NaN in cashflow, IRR calculation cannot be performed: ' + str(cashflow))
            return 0
        return numpy_financial.irr(cashflow)

    def __net_present_value(self, revenue: numpy.ndarray) -> float:
        return self.__npv(self.__investment_costs, revenue, self.__discount_rate)

    def __profitability_index(self, revenue: numpy.ndarray) -> float:
        return self.__npv(0, revenue, self.__discount_rate) / self.__investment_costs

    def __return_on_investment(self, revenue: numpy.ndarray) -> float:
        return self.__net_present_value(revenue) / self.__investment_costs

    def __levelized_cost_of_storage_energy(self) -> float:
        energy: numpy.ndarray = self.get_data().discharge_energy_per_year / 1000.0  # conversion to MWh
        energy_per_year: float = self.__npv(0, energy, self.__discount_rate)
        if energy_per_year:
            # TODO include further (discounted) costs for LCOS
            return self.__investment_costs / energy_per_year
        else:
            return numpy.nan

    @staticmethod
    def __npv(invest: float, cashflow: numpy.ndarray, discount_rate: float) -> float:
        cashflow = numpy.append(numpy.array([-invest]), cashflow)
        return numpy_financial.npv(discount_rate, cashflow)

    def __get_assumptions(self) -> list:
        assumptions: list = list()
        assumptions.append(EvaluationResult(Description.Economical.INVESTMENT_COSTS, Unit.EURO, self.__investment_costs))
        assumptions.append(EvaluationResult(Description.Economical.DISCOUNT_RATE, Unit.PERCENTAGE, self.__discount_rate * 100))
        for revenue_stream in self.__revenue_streams:
            assumptions.extend(revenue_stream.get_assumptions())
        return assumptions

    def print_results(self):
        if not self._print_to_console:
            return
        print('\n\n[ECONOMIC ANALYSIS]:')
        print('ASSUMPTIONS:')
        for assumption in self.__get_assumptions():
            print(assumption.to_console())
        print('\nRESULTS:')
        for evaluation_result in self.evaluation_results:
            print(evaluation_result.to_console())
        self.__print_warning()

    def __print_warning(self):
        time: numpy.ndarray = self.get_data().time
        start: float = time[0]
        end: float = time[-1]
        duration: float = end - start
        date = datetime.fromtimestamp(start)
        year_end: float = date.replace(year=date.year + 1).timestamp()
        year: float = year_end - start
        if duration < year:
            print('Note: Simuation time is less than one year (' + format_float(duration / year) +
                  ' a). Results only cover the simulation time.')

    def close(self) -> None:
        self.__log.close()
        for revenue_stream in self.__revenue_streams:
            revenue_stream.close()



import csv
import gzip
import threading
import time
from datetime import datetime
from queue import Queue, Empty
import numpy
import pandas as pd
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.data.data_handler import DataHandler
from simses.commons.log import Logger
from simses.commons.state.abstract_state import State
from simses.commons.utils.utilities import remove_all_files_from, copy_all_files, create_directory_for, \
    all_non_abstract_subclasses_of


class CSVDataHandler(threading.Thread, DataHandler):
    """
    Export for Battery data during the simulation into one CSV file_name.
    Batteries send their parameters for every timestamp to this class, which adds them to the CSV file_name.
    For Multiprocessing a queue is used.
    """

    USE_GZIP_COMPRESSION: bool = True
    EXT_CSV: str = '.csv'
    # Other compression formats: zip, bz2, xz
    COMP: str = 'gzip'
    EXT_COMP: str = '.gz'

    def __init__(self, result_dir: str, config: GeneralSimulationConfig, export_cls=State):
        super().__init__()
        self.__log: Logger = Logger(type(self).__name__)
        self.__queue = Queue()
        self.__append = True
        self.__simulation_running = True
        self.__result_folder: str = result_dir + 'EES_' + datetime.now().strftime('%Y%m%dT%H%M%SM%f')
        create_directory_for(self.__result_folder, warn=True)
        self.__working_directory = result_dir
        remove_all_files_from(self.__working_directory)
        self.__classes: list = all_non_abstract_subclasses_of(export_cls)
        # Deactivating export interval to prohibit wrong analysis results
        # self.__export_interval: float = config.export_interval
        self.__export_interval: float = 1
        self.__timestep: float = config.timestep

    def run(self) -> None:
        """
        Function to write data from queue to csv.

        Parameters
        ----------

        Returns
        -------

        """
        tstmps: dict = dict()
        file_writer: dict = dict()
        file_streams: list = list()
        # for cls in self.__classes:
        #     self.__register_writer(cls, file_writer, tstmps, file_streams)
        while self.__simulation_running or not self.__queue.empty():
            try:
                data: dict = self.__queue.get_nowait()
                self.__log.debug('Write data from queue to output file')
                for key in list(data.keys()):
                    # print(data[key])
                    if key not in file_writer.keys():
                        for cls in self.__classes:
                            if cls.__name__ == key:
                                self.__register_writer(cls, file_writer, tstmps, file_streams)
                    ts = data[key]['TIME']
                    id = data[key]['ID']
                    if id not in tstmps[key]:
                        tstmps[key][id] = 0
                    if ts >= tstmps[key][id] + self.__export_interval * self.__timestep:
                        values: list = data[key]['DATA']
                        if numpy.isnan(values).any():
                            self.__log.error('State ' + key + ' has NaN: ' + str(values))
                        file_writer[key].writerow(values)
                        tstmps[key][id] = ts
            except Empty:
                self.__log.debug("Empty Queue")
                time.sleep(0.01)
        self.__log.info('Write to output file finished')
        for stream in file_streams:
            stream.close()
        file_writer.clear()
        file_streams.clear()

    def __register_writer(self, cls, file_writer, tstmps, file_streams) -> None:
        self.__log.info('Registering ' + cls.__name__)
        file_name = self.__working_directory + cls.__name__ + self.EXT_CSV
        if self.USE_GZIP_COMPRESSION:
            file_stream = gzip.open(file_name + self.EXT_COMP, 'at' if self.__append else 'wt', newline='')
        else:
            file_stream = open(file_name, 'a' if self.__append else 'w', newline='')
        writer = csv.writer(file_stream)
        writer.writerow(cls.header())
        file_writer[cls.__name__] = writer
        tstmps[cls.__name__] = {}
        file_streams.append(file_stream)

    def transfer_data(self, data: dict) -> None:
        """
        Function to place data into the queue. This data is picked up by the run function and written into a CSV file_name.

        Parameters
        ----------
        data : list

        Returns
        -------

        """
        self.__queue.put(data)

    def simulation_done(self) -> None:
        """
        Function to let the DataExport Thread know that the simulation is done and stop after emptying the queue.

        Parameters
        ----------

        Returns
        -------

        """
        self.__simulation_running = False

    def copy_results_to_destination(self):
        """
        Transfer the resulting csv file_name as well as the configuration file_name to the destination folder defined in the
        class creation.

        Returns
        -------

        """
        copy_all_files(self.__working_directory, self.__result_folder)
        remove_all_files_from(self.__working_directory)

    def close(self) -> None:
        """
        Closing all open resources in data export

        Parameters
        ----------

        Returns
        -------

        """

        self.__log.debug('Simulation done')
        self.simulation_done()
        self.__log.debug('Waiting for export thread to finish')
        if self.is_alive():
            self.join()
        self.copy_results_to_destination()
        self.__log.close()

    @classmethod
    def get_data_from(cls, path: str, state_cls) -> pd.DataFrame:
        """
        Reads file for Class in path as pandas dataframe

        Parameters
        ----------
        path : Path of simulation results
        state_cls : State class

        Returns
        -------
        pandas.Dataframe:
            Data of file in path

        """
        filename: str = path + '/' + state_cls.__name__ + cls.EXT_CSV
        if cls.USE_GZIP_COMPRESSION:
            data: pd.DataFrame = pd.read_csv(filename + cls.EXT_COMP, compression=cls.COMP, header=0)
        else:
            data: pd.DataFrame = pd.read_csv(filename, header=0)
        # print('==== ' + state_cls.__name__ + ' ====')
        for key in data.keys():
            data[key] = pd.to_numeric(data[key], errors='coerce')
        # data = data.replace(np.nan, 0, regex=True)
        # print(data.dtypes)
        return data



from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.state.technology.storage import StorageTechnologyState
from .cycle_detector import CycleDetector
from ..config.simulation.battery import BatteryConfig


class RainflowCycleDetector(CycleDetector):
    """
    Continious Rainflow Counting based on Xu et al.
    https://doi.org/10.1109/TPWRS.2017.2733339
    and Amzallag et al.
    https://doi.org/10.1016/0142-1123(94)90343-3
    """

    def __init__(self, start_soc: float, general_config: GeneralSimulationConfig, battery_config: BatteryConfig):
        super().__init__()
        self.__flag_charging = True
        self.__depth_of_cycle: float = 0.0
        self.__full_equivalent_cycles: float = 0.0
        self.__c_rate: float = 0.0
        self.__mean_soc: float = 0.0

        self.__last_soc = start_soc
        self.__start_soc = start_soc

        self.__soc  = [self.__start_soc]
        self.__time = [general_config.start]
        self.__end_time = general_config.end - general_config.timestep
        self.__soc_lower_limit = battery_config.min_soc
        self.__soc_upper_limit = battery_config.max_soc

    def cycle_detected(self, time: float, state: StorageTechnologyState) -> bool:

        soc = round(state.soc, 4)

        ### Spezific cases:
        # last cycle
        if time > self.__end_time:
            self.__soc.append(soc)
            self.__time.append(state.time)
            return self.__last_cycle()
        # no change of SOC
        elif self.__soc[-1] == soc:
            self.__time[-1] = state.time
            return False
        # start of cycle: Set flag charging or discharging
        elif len(self.__soc) == 1:
            self.__initial_event(soc=soc, time=state.time)
            return False
        # check if sign changes, otherwise overwrite last value
        elif self.__check_event(soc=soc, time=state.time):
            self.__soc.append(soc)
            self.__time.append(state.time)

        # cycle at 0-x-0 or 1-x-1
        if len(self.__soc) == 3 and self.__cycle_at_limit():
            return True
        # no cycle
        elif len(self.__soc) < 4:
            return False

        ### Finally RainFlow Algorithm
        delta_soc = []
        soc_preceded = self.__soc[0]
        for soc in self.__soc[-3:]:
            delta_soc.append(abs(soc_preceded - soc))
            soc_preceded = soc

        if delta_soc[1] > delta_soc [0] or delta_soc[1] > delta_soc[2]:
            return False
        else:
            self._update_cycle_steps(soc, time)
            return True

    def get_depth_of_cycle(self) -> float:
        return self.__depth_of_cycle

    def get_delta_full_equivalent_cycle(self) -> float:
        return self.__depth_of_cycle

    def get_crate(self) -> float:
        return self.__c_rate

    def get_mean_soc(self) -> float:
        return self.__mean_soc

    def get_full_equivalent_cycle(self) -> float:
        return self.__full_equivalent_cycles

    def reset(self) -> None:
        # TODO is this really needed?
        self.__depth_of_cycle = 0
        self.__full_equivalent_cycles = 0

    def _update_cycle_steps(self, soc: float, time_passed: float) -> None:
        """
        updates all values within the cycle detector, if no cycle was detected, but the SOC changed

        Returns
        -------

        """
        self.__depth_of_cycle = abs(self.__soc[-2] - self.__soc[-3])
        self.__full_equivalent_cycles += self.__depth_of_cycle
        self.__mean_soc = (self.__soc[-3] + self.__soc[-2]) / 2
        self.__c_rate = abs(self.__soc[-2] - self.__soc[-3]) / (abs(self.__time[-2] - self.__time[-3]) +
                                                                abs(self.__time[-1] - self.__time[-2]) / 2 )

        self.__soc.pop(-3)
        self.__soc.pop(-2)
        self.__time.pop(-3)
        self.__time.pop(-2)

    def __initial_event(self, soc: float, time: float) -> None:
        if soc > self.__soc[0]:
            self.__flag_charging = True
        else:
            self.__flag_charging = False
        self.__soc.append(soc)
        self.__time.append(time)

    def __check_event(self, soc: float, time: float) -> bool:
        if soc < self.__soc[-1] and not self.__flag_charging:
            self.__soc[-1] = soc
            self.__time[-1] = time
            return False
        elif soc < self.__soc[-1]:
            self.__flag_charging = False
            return True
        elif soc > self.__soc[-1] and self.__flag_charging:
            self.__soc[-1] = soc
            self.__time[-1] = time
            return False
        else:
            self.__flag_charging = True
            return True

    def __last_cycle(self) -> bool:
        if len(self.__soc) == 2 and self.__soc[0] == self.__soc[1]: # no cycle
            return False
        elif len(self.__soc) == 2: # only half cycle
            self.__depth_of_cycle = abs(self.__soc[1] - self.__soc[0])
            self.__full_equivalent_cycles += self.__depth_of_cycle / 2
            self.__mean_soc = (self.__soc[0] + self.__soc[1]) / 2
            half_cycle_time = abs(self.__time[0] - self.__time[1])
            self.__c_rate = self.__depth_of_cycle / half_cycle_time
            self.__soc.clear()
            self.__time.clear()
            return True
        else:
            self.__depth_of_cycle = abs(self.__soc[-2] - self.__soc[-3])
            self.__full_equivalent_cycles += self.__depth_of_cycle
            self.__mean_soc = (self.__soc[-3] + self.__soc[-2]) / 2
            half_cycle_time = (abs(self.__time[-1] - self.__time[-2]) + abs(self.__time[-2] - self.__time[-3])) / 2
            self.__c_rate = abs(self.__soc[-2] - self.__soc[-3]) / half_cycle_time
            self.__soc.clear()
            self.__time.clear()
            return True

    def __cycle_at_limit(self) -> bool:
        if (self.__soc[0] == self.__soc_lower_limit and self.__soc[-1] == self.__soc_lower_limit) or \
                (self.__soc[0] == self.__soc_upper_limit and self.__soc[-1] == self.__soc_upper_limit):

            if self.__soc[-1] == self.__soc_lower_limit:
                self.__flag_charging = True
            else:
                self.__flag_charging = False

            self.__depth_of_cycle = abs(self.__soc[-2] - self.__soc[-3])
            self.__full_equivalent_cycles += self.__depth_of_cycle
            self.__mean_soc = (self.__soc[-3] + self.__soc[-2]) / 2
            half_cycle_time = (abs(self.__time[-1] - self.__time[-2]) + abs(self.__time[-2] - self.__time[-3])) / 2
            self.__c_rate = abs(self.__soc[-2] - self.__soc[-3]) / half_cycle_time
            self.__soc = [self.__soc[-1]]
            self.__time = [self.__time[-1]]
            return True
        else:
            return False


import math
import pandas as pd
import scipy.interpolate
from simses.commons.config.data.battery import BatteryDataConfig
from simses.commons.config.simulation.battery import BatteryConfig
from simses.commons.log import Logger
from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.technology.lithium_ion.cell.electric.properties import ElectricalCellProperties
from simses.technology.lithium_ion.cell.format.abstract import CellFormat
from simses.technology.lithium_ion.cell.format.round_26650 import RoundCell26650
from simses.technology.lithium_ion.cell.thermal.properties import ThermalCellProperties
from simses.technology.lithium_ion.cell.type import CellType


class SonyLFP(CellType):
    """
    Source SONY_US26650FTC1_Product Specification and Naumann, Maik. Techno-economic evaluation of stationary
    lithium_ion energy storage systems with special consideration of aging.
    PhD Thesis. Technical University Munich, 2018.
    """

    __SOC_HEADER: str = 'SOC'
    __SOC_IDX: int = 0
    __TEMP_IDX: int = 1
    __C_RATE_IDX: int = 0
    __ETA_IDX: int = 1

    __CELL_VOLTAGE: float = 3.2  # V
    __CELL_CAPACITY: float = 3.0  # Ah
    __MAX_VOLTAGE: float = 3.6  # V
    __MIN_VOLTAGE: float = 2.0  # V
    __MAX_CHARGE_RATE: float = 1.0  # 1/h
    __MAX_DISCHARGE_RATE: float = 6.6  # 1/h
    __SELF_DISCHARGE_RATE: float = 0.0 / (365 / 12)  # X.X%-soc per day, e.g., 0.1 % per month
    # __SELF_DISCHARGE_RATE: float = 0.1 / (365 / 12)  # X.X%-soc per day, e.g., 0.1 % per month
    __COULOMB_EFFICIENCY: float = 1.0  # p.u.

    __MIN_TEMPERATURE: float = 273.15  # K
    __MAX_TEMPERATURE: float = 333.15  # K
    __SPECIFIC_HEAT: float = 1001  # J/kgK
    __CONVECTION_COEFFICIENT: float = 15  # W/m2K
    __MASS: float = 0.07  # kg per cell

    __ELECTRICAL_PROPS: ElectricalCellProperties = ElectricalCellProperties(__CELL_VOLTAGE, __CELL_CAPACITY,
                                                                            __MIN_VOLTAGE, __MAX_VOLTAGE,
                                                                            __MAX_CHARGE_RATE, __MAX_DISCHARGE_RATE,
                                                                            __SELF_DISCHARGE_RATE, __COULOMB_EFFICIENCY)
    __THERMAL_PROPS: ThermalCellProperties = ThermalCellProperties(__MIN_TEMPERATURE, __MAX_TEMPERATURE, __MASS,
                                                                   __SPECIFIC_HEAT, __CONVECTION_COEFFICIENT)
    __CELL_FORMAT: CellFormat = RoundCell26650()

    def __init__(self, voltage: float, capacity: float, soh: float, battery_config: BatteryConfig,
                 battery_data_config: BatteryDataConfig):
        super().__init__(voltage, capacity, soh, self.__ELECTRICAL_PROPS, self.__THERMAL_PROPS, self.__CELL_FORMAT,
                         battery_config)
        self.__log: Logger = Logger(type(self).__name__)
        rint_file: str = battery_data_config.lfp_sony_rint_file
        internal_resistance = pd.read_csv(rint_file)  # Ohm
        # SOC array has only 11 entries 0:0.1:1
        soc_arr = internal_resistance.iloc[:, self.__SOC_IDX]
        # Temperature array has only 4 values (283.15 K, 298.15 K, 313.15 K, 333.15 K)
        temp_arr = internal_resistance.iloc[:4, self.__TEMP_IDX]
        # internal resistance for charge, Column 2:5 for charging
        rint_mat_ch = internal_resistance.iloc[:, 2:6]
        # internal resistance for discharge, Column 6:9 for charging
        rint_mat_dch = internal_resistance.iloc[:, 6:]
        self.__rint_ch_interp2d = scipy.interpolate.interp2d(temp_arr.T, soc_arr, rint_mat_ch, kind='linear')
        self.__rint_dch_interp2d = scipy.interpolate.interp2d(temp_arr.T, soc_arr, rint_mat_dch, kind='linear')

    def get_open_circuit_voltage(self, battery_state: LithiumIonState) -> float:
        '''Parameters build with ocv fitting'''
        a1 = -116.2064
        a2 = -22.4512
        a3 = 358.9072
        a4 = 499.9994
        b1 = -0.1572
        b2 = -0.0944
        k0 = 2.0020
        k1 = -3.3160
        k2 = 4.9996
        k3 = -0.4574
        k4 = -1.3646
        k5 = 0.1251

        soc = battery_state.soc

        ocv = k0 + \
              k1 / (1 + math.exp(a1 * (soc - b1))) + \
              k2 / (1 + math.exp(a2 * (soc - b2))) + \
              k3 / (1 + math.exp(a3 * (soc - 1))) +\
              k4 / (1 + math.exp(a4 * soc)) +\
              k5 * soc

        return ocv * self.get_serial_scale()

    def get_internal_resistance(self, battery_state: LithiumIonState) -> float:
        if battery_state.is_charge:
            rint = self.__rint_ch_interp2d(battery_state.temperature, battery_state.soc)
        else:
            rint = self.__rint_dch_interp2d(battery_state.temperature, battery_state.soc)
        return float(rint) / self.get_parallel_scale() * self.get_serial_scale()

    def close(self):
        self.__log.close()



from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.profile.file import FileProfile
from simses.commons.profile.power.power_profile import PowerProfile


class FilePowerProfile(PowerProfile):
    """
    FilePowerProfile is a generic implementation of a PowerProfile. It provides power values from a file using
    FileProfile library.
    """

    class Header:
        ANNUAL_CONSUMPTION: str = 'Annual load consumption in kWh'
        PEAK_POWER: str = 'Nominal power in kWp'

    def __init__(self, config: GeneralSimulationConfig, filename: str, delimiter: str = ',', scaling_factor: float = 1.0):
        """
        Constructor of FilePowerProfile

        Parameters
        ----------
        config :
            simulation configuration
        filename :
            path to file
        delimiter :
            delimter of values in file, default: ,
        scaling_factor :
            linear scaling of values, default: 1
        """
        super().__init__()
        self.__file: FileProfile = FileProfile(config, filename, delimiter, scaling_factor)

    def next(self, time: float) -> float:
        return self.__file.next(time)

    def profile_data_to_list(self, sign_factor=1) -> [float]:
        """
        Extracts the whole time series as a list and resets the pointer of the (internal) file afterwards

        Parameters
        ----------
        sign_factor :

        Returns
        -------
        list:
            profile values as a list

        """
        time, values = self.__file.profile_data_to_list(sign_factor)
        return values

    def close(self):
        self.__file.close()



import os
from configparser import ConfigParser
from datetime import datetime, timezone

import pytest

from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.profile.power.file import FilePowerProfile
from simses.commons.state.system import SystemState
from simses.commons.utils.utilities import remove_file
from simses.logic.energy_management.strategy.basic.residential_pv_greedy import \
    ResidentialPvGreedy

DELIMITER: str = ','
HEADER: str = '# Unit: W'
system_state: SystemState = SystemState(0,0)
START: int = 0
END: int = 4
STEP: int = 1
FILE_NAME_LOAD: str = os.path.join(os.path.dirname(__file__), 'mockup_power_load_profile.csv')
FILE_NAME_PV: str = os.path.join(os.path.dirname(__file__), 'mockup_power_pv_profile.csv')


def create_load_file():
    with open(FILE_NAME_LOAD, mode='w') as file:
        file.write(HEADER + '\n')
        value = START
        tstmp = START
        while tstmp <= END:
            file.write(str(tstmp) + DELIMITER + str(value) + '\n')
            tstmp += STEP
            value += 1

    power_load_profile = FilePowerProfile(config=create_general_config(), filename=FILE_NAME_LOAD,
                                          delimiter=DELIMITER, scaling_factor=1)
    return power_load_profile

def create_pv_file():
    with open(FILE_NAME_PV, mode='w') as file:
        file.write(HEADER + '\n')
        value = END
        tstmp = START
        while tstmp <= END:
            file.write(str(tstmp) + DELIMITER + str(value) + '\n')
            tstmp += STEP
            value -= 1

    power_pv_profile = FilePowerProfile(config=create_general_config(), filename=FILE_NAME_PV,
                                          delimiter=DELIMITER, scaling_factor=1)
    return power_pv_profile

def create_general_config():
    conf: ConfigParser = ConfigParser()
    conf.add_section('GENERAL')
    date = datetime.fromtimestamp(START, timezone.utc)
    conf.set('GENERAL', 'START', date.strftime("%Y-%m-%d %H:%M:%S"))
    date = datetime.fromtimestamp(END, timezone.utc)
    conf.set('GENERAL', 'END', date.strftime("%Y-%m-%d %H:%M:%S"))
    conf.set('GENERAL', 'TIME_STEP', str(STEP))
    return GeneralSimulationConfig(config=conf)

@pytest.fixture(scope='module')
def uut():
    power_load_profile = create_load_file()
    power_pv_profile = create_pv_file()
    uut = ResidentialPvGreedy(power_load_profile, power_pv_profile)
    yield uut
    uut.close()
    remove_file(FILE_NAME_LOAD)
    remove_file(FILE_NAME_PV)

# time = power
@pytest.mark.parametrize('time, soc, result',
                         [
                            (0, 0, 4),
                            (1, 0, 2),
                            (2, 0, 0),
                            (3, 0, -2),
                            (4, 0, -4)
                         ]
                         )
def test_next(time, soc, result, uut):
    system_state.soc = soc
    assert abs(uut.next(time, system_state) - result) <= 1e-10


from configparser import ConfigParser

from simses.commons.config.data.data_config import DataConfig


class FuelCellDataConfig(DataConfig):

    def __init__(self, path: str = None, config: ConfigParser = None):
        super().__init__(path, config)
        self.__section: str = 'FUEL_CELL_DATA'

    @property
    def fuel_cell_data_dir(self) -> str:
        """Returns directory of fuelcell data files"""
        return self.get_data_path(self.get_property(self.__section, 'FUEL_CELL_DATA_DIR'))

    @property
    def pem_fuel_cell_pc_file(self) -> str:
        """Returns filename for polarisation curve for PEM-Fuelcell"""
        return self.fuel_cell_data_dir + self.get_property(self.__section, 'PEM_FUEL_CELL_PC_FILE')

    @property
    def pem_fuel_cell_power_file(self) -> str:
        """Returns filename for power curve for PEM-Fuelcell"""
        return self.fuel_cell_data_dir + self.get_property(self.__section, 'PEM_FUEL_CELL_POWER_FILE')

    @property
    def jupiter_fuel_cell_pc_file(self) -> str:
        """Returns filename for polarisation curve for Jupiter-Fuelcell"""
        return self.fuel_cell_data_dir + self.get_property(self.__section, 'JUPITER_FUEL_CELL_PC_FILE')

    @property
    def jupiter_fuel_cell_power_file(self) -> str:
        """Returns filename for power curve for Jupiter-Fuelcell"""
        return self.fuel_cell_data_dir + self.get_property(self.__section, 'JUPITER_FUEL_CELL_POWER_FILE')



from simses.commons.config.simulation.electrolyzer import ElectrolyzerConfig
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.state.technology.electrolyzer import ElectrolyzerState
from simses.technology.hydrogen.electrolyzer.degradation.calendar.pem_analytic import \
    CalendarDegradationPemElMultiDimAnalyitic
from simses.technology.hydrogen.electrolyzer.degradation.cyclic.pem_analytic import \
    CyclicDegradationPemElMultiDimAnalytic
from simses.technology.hydrogen.electrolyzer.degradation.degradation_model import \
    DegradationModel
from simses.technology.hydrogen.electrolyzer.stack.stack_model import ElectrolyzerStackModel


class PemElMultiDimAnalyticDegradationModel(DegradationModel):

    def __init__(self, electrolyzer: ElectrolyzerStackModel, config: ElectrolyzerConfig,
                 general_config: GeneralSimulationConfig):
        super().__init__(CyclicDegradationPemElMultiDimAnalytic(),
                         CalendarDegradationPemElMultiDimAnalyitic(general_config))

        self.__end_of_life = config.eol
        self.__rev_voltage_bol = electrolyzer.get_reference_voltage_eol(0.0, 1.0)
        self.__voltage_increase_eol = 0.27  # V
        self.__soh = 1.0  # p.u.

    def calculate_soh(self, state: ElectrolyzerState) -> None:
        self.__soh = 1.0 - 0.2 * (state.reference_voltage - self.__rev_voltage_bol) / self.__voltage_increase_eol

    def get_soh_el(self):
        return self.__soh


from simses.system.auxiliary.auxiliary import Auxiliary
from simses.system.power_electronics.dcdc_converter.abstract_dcdc_converter import DcDcConverter


class NoDcDcConverter(DcDcConverter):

    """
    This class acts like there is no DCDC converter included in the systems. It passes the voltage of the storage
    system to the intermediate circuit (which is in this case not an intermediate circuit anymore). Multiple parallel
    connected NoDcDcConverter produce invalid results.
    """

    __volumetric_power_density = 143 * 10 ** 6  # W / m3
    __GRAVIMETRIC_POWER_DENSITY = 11000  # W/kg
    __SPECIFIC_SURFACE_AREA = 0.0001  # in m2 / W  # TODO add exact values

    def __init__(self, intermediate_circuit_voltage: float):
        super(NoDcDcConverter, self).__init__(intermediate_circuit_voltage)
        self.__dc_power_loss: float = 0
        self.__dc_power: float = 0
        self.__dc_current: float = 0
        self.__dc_circuit_voltage: float = intermediate_circuit_voltage

    def calculate_dc_current(self, dc_power: float, storage_voltage: float) -> None:
        self.__dc_power = dc_power
        self.__dc_circuit_voltage = storage_voltage
        self.__dc_current = dc_power / storage_voltage

    def reverse_calculate_dc_current(self, dc_power: float, storage_voltage: float) -> None:
        self.__dc_power = dc_power
        self.__dc_circuit_voltage = storage_voltage
        self.__dc_current = dc_power / storage_voltage

    @property
    def max_power(self) -> float:
        return 1e12

    @property
    def dc_power_loss(self):
        return self.__dc_power_loss

    @property
    def dc_power(self):
        return self.__dc_power

    @property
    def intermediate_circuit_voltage(self) -> float:
        return self.__dc_circuit_voltage

    @property
    def volume(self) -> float:
        # return self.max_power / self.__volumetric_power_density
        return 0

    @property
    def mass(self):
        return 0

    @property
    def surface_area(self) -> float:
        return 0

    @property
    def dc_current(self):
        return self.__dc_current

    def get_auxiliaries(self) -> [Auxiliary]:
        return list()

    def close(self) -> None:
        pass



from configparser import ConfigParser
from simses.commons.config.abstract_config import Config
from simses.commons.utils.utilities import get_path_for


class SimulationConfig(Config):
    """
    All simulation configs are inherited from this class
    """

    CONFIG_NAME: str = 'simulation'
    CONFIG_PATH: str = get_path_for(CONFIG_NAME + Config.DEFAULTS)

    def __init__(self, path: str, config: ConfigParser):
        if path is None:
            path = self.CONFIG_PATH
        super().__init__(path, self.CONFIG_NAME, config)


def create_dict_from(properties: [str], delimiter: str = ',') -> dict:
    res: dict = dict()
    for prop in properties:
        items: list = prop.split(delimiter)
        name: str = items.pop(0)
        if name in res.keys():
            raise Exception(name + ' is not unique. Please check your config file.')
        res[name] = items
    return res


def create_list_from(properties: [str], delimiter: str = ',') -> [[str]]:
    res: [[str]] = list()
    for prop in properties:
        res.append(prop.split(delimiter))
    return res


def clean_split(properties: str, delimiter: str = '\n') -> [str]:
    props: [str] = properties.replace(' ', '').replace('\t', '').split(delimiter)
    copy = props[:]
    for prop in copy:
        if not prop:
            props.remove(prop)
    return props



import os
import threading
from configparser import ConfigParser
from multiprocessing import Queue
from os.path import dirname
from simses.analysis.storage import StorageAnalysis
from simses.commons.config.log import LogConfig
from simses.commons.state.system import SystemState
from simses.commons.utils.utilities import create_directory_for
from simses.simulation.simulator import StorageSimulation


class SimSES(threading.Thread):

    """
    SimSES is a simulation and analysis tool for complex energy storage systems. The tool can be used via the run method
    (or alternatively as a Thread with its start method). The simulation and analysis can be run separately, the
    analysis takes the results of the last run from the configured simulation.

    SimSES can be configured via the INI files in config. DEFAULT configs (*.defaults.ini) can be overwritten by
    LOCAL configs (*.local.ini). Furthermore, the configs can be overwritten by a ConfigParser passed to the SimSES
    constructor. This is especially useful for sensitivity analysis.

    The processing package provides a functionality for multiple simulations and analysis by using all available cores.
    For this purpose, python's multiprocessing package is used. An example as well as a readme is provided in the
    processing package how to use and configure it.

    SimSES can also be used directly in other tools by providing the run_one_simulation_step and
    evaluate_multiple_simulation_steps methods. With these methods SimSES can be integrated in other simulation
    frameworks acting as a storage system. SimSES needs to be closed manually after the simulation from outer scope is
    completed.
    """

    __DIR: str = dirname(__file__)
    __VERSION_FILE: str = __DIR + '/VERSION'
    __GIT_FILE: str = __DIR + '/../.git/FETCH_HEAD'

    def __init__(self, path: str, name: str, do_simulation: bool = True, do_analysis: bool = True,
                 simulation_config: ConfigParser = None, analysis_config: ConfigParser = None, queue: Queue = None,
                 batch_dir: str = 'batch/'):
        """
        Constructor of SimSES

        Parameters
        ----------
        path :
            absolute path where to store results
        name :
            simulation name (will be concatenated with path to a unique path)
        do_simulation :
            flag for allowing or prohibiting execution of the simulation
        do_analysis :
            flag for allowing or prohibiting execution of the analysis
        simulation_config :
            ConfigParser overwriting configuration provided by INI files for simulation
        analysis_config :
            ConfigParser overwriting configuration provided by INI files for analysis
        queue :
            PrinterQueue provides progess information for the user in case of using the processing package
        batch_dir :
            Relative path of directory for comparison of results from multiple simulations using the processing package
        """
        super().__init__()
        self.__do_simulation = do_simulation
        self.__do_analysis = do_analysis
        self.__name: str = name
        batch_dir = path + batch_dir
        self.__path = path + name + '/'
        if self.__do_simulation:
            create_directory_for(path)
            self.__storage_simulation: StorageSimulation = StorageSimulation(self.__path, simulation_config, queue)
        if self.__do_analysis:
            self.__storage_analysis: StorageAnalysis = StorageAnalysis(self.__path, analysis_config, batch_dir, self.version)

    @property
    def name(self) -> str:
        """
        Returns
        -------
        str:
            string representation of the simulation name
        """
        return self.__name

    @property
    def version(self) -> str:
        """
        Returns
        -------
        str:
            current version of simses including, if available, git commit hash (short version) from last fetch
        """
        version = self.__get_version_simses()
        version += self.__get_version_git()
        return version

    def __get_version_git(self):
        version: str = ''
        try:
            with open(self.__GIT_FILE, 'r') as file:
                version += '-' + file.readline().split('\t')[0].rstrip()[:8]
        except FileNotFoundError:
            pass
        return version

    def __get_version_simses(self):
        version: str = ''
        try:
            with open(self.__VERSION_FILE, 'r') as file:
                version += file.readline().rstrip()
        except FileNotFoundError:
            version = 'unknown'
        return version

    def run(self) -> None:
        """
        Runs the configured simulation and analysis and closes afterwards
        """
        self.run_simulation()
        self.run_analysis()
        self.close()

    def run_one_simulation_step(self, time: float, power: float = None, power_dist: [float] = None, adapted_time: float = None) -> None:
        """
        Runs only one step of the simulation with the given time and power. The system is configured as mentioned
        in the class description.

        If no power value is provided, the configured energy management system will provide a power value for
        the given time.

        Calculated values can be received via the state property. It provides information about the whole
        storage system, e.g. SOC.

        Parameters
        ----------
        time :
            epoch timestamp in s
        power :
            power value in W
        power_dist:
            power value in W for every AC system (optional external power distribution logic)
        adapted_time: time offset in seconds to enable looping
        """
        if adapted_time is not None:
            self.__storage_simulation.run_one_step(ts=time, power=power, power_dist=power_dist, ts_adapted=adapted_time)
        else:
            self.__storage_simulation.run_one_step(ts=time, power=power, power_dist=power_dist)

    def evaluate_multiple_simulation_steps(self, start: float, timestep: float, power: list) -> [SystemState]:
        """
        Runs multiple steps of the simulation with the given start time, timestep and power list. The system is
        configured as mentioned in the class description.

        If no power list is provided, the simulation will not be advanced.

        Calculated values will be returned. They provide information about the whole storage system, e.g. SOC,
        for each timestep.

        Parameters
        ----------
        start :
            start time in s
        timestep :
            timestep in s
        power :
            list of power for each timestep in W

        Returns
        ----------
        list:
            Returns a list of system states for each timestep
        """
        return self.__storage_simulation.evaluate_multiple_steps(start, timestep, power)

    def run_simulation(self) -> None:
        """
        Runs only the simulation as configured, if allowed
        """
        if self.__do_simulation:
            self.__storage_simulation.run()

    def run_analysis(self) -> None:
        """
        Runs only the analysis as configured, if allowed
        """
        if self.__do_analysis:
            self.__storage_analysis.run()

    @property
    def state(self) -> SystemState:
        """
        Usage only supported in combination with run_one_simulation_step

        Returns
        -------
        SystemState:
            current state of the system providing information of calculated results
        """
        return self.__storage_simulation.state

    @property
    def ac_system_states(self) -> [SystemState]:
        """
        Usage only supported in combination with run_one_simulation_step

        Returns
        -------
        [SystemState]:
            current states of all AC systems providing information of calculated results
        """
        return self.__storage_simulation.ac_system_states

    def reset_profiles(self, ts_adapted: float) -> None:
        """
        Enables looping of the simulation beyond the original length of the time series for the AmbientThermalModel and
        SolarIrradiationModel
        """
        self.__storage_simulation.reset_profiles(ts_adapted)

    def close(self) -> None:
        """
        Closes all resources of simulation and analysis
        """
        self.close_simulation()
        self.close_analysis()

    def close_simulation(self) -> None:
        """
        Closes all resources of simulation
        """
        if self.__do_simulation:
            self.__storage_simulation.close()

    def close_analysis(self) -> None:
        """
        Closes all resources of analysis
        """
        if self.__do_analysis:
            self.__storage_analysis.close()

    @classmethod
    def set_log_config(cls, configuration: ConfigParser) -> None:
        """
        Class method for setting the global log configuration

        Parameters
        ----------
        configuration :
            ConfigParser will overwrite global log configuration
        """
        LogConfig.set_config(configuration)


if __name__ == "__main__":
    # minimum working example
    # config_generator: SimulationConfigGenerator = SimulationConfigGenerator()
    # config_generator.set_simulation_time('2014-01-01 00:00:00', '2014-02-01 00:00:00')
    # config: ConfigParser = config_generator.get_config()
    path: str = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    result_path: str = os.path.join(path, 'results').replace('\\','/') + '/'
    simulation_name: str = 'simses_1'
    simses: SimSES = SimSES(result_path, simulation_name, do_simulation=True, do_analysis=True)
    # simses: SimSES = SimSES(result_path, simulation_name, do_simulation=True, do_analysis=True, simulation_config=config)
    simses.start()
    simses.join()



from simses.commons.cycle_detection.cycle_detector import CycleDetector
from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.technology.lithium_ion.cell.type import CellType
from simses.technology.lithium_ion.degradation.cyclic.cyclic_degradation import CyclicDegradationModel

import numpy as np


class NCAGenericCyclicDegradationModel(CyclicDegradationModel):

    def __init__(self, cell_type: CellType, cycle_detector: CycleDetector):
        super().__init__(cell_type, cycle_detector)
        self.__capacity_loss = 0
        self.__initial_capacity = self._cell.get_nominal_capacity()

        # parameters
        self.__z = 0.172396498182181
        self.__a = -6.4598243817573E-07
        self.__b = 0.000159249929508653
        self.__c = 0.0162972227060786
        self.__d = 1.22012389207326
        self.__e = 0.291743390220046
        self.__f = 0.29174367528008
        self.__g = 0.291694646922772
        self.__h = 0.291609580760524
        self.__i = 0.000440503468592159
        self.__j = 0.235399850460447
        self.__k = 1
        self.__l = 1
        self.__m = 0.0927398171669826
        self.__n = 1

        # constants
        self.__R = 8.314462  # universal gas constant

    def fec_dependency(self, fec):
        # degradation depends on fec as a power law function
        return fec**self.__z

    def temperature_dependency(self, temp):
        # degradation depends on temperature as a cubic function
        return self.__a * temp ** 3 + self.__b * temp ** 2 + self.__c * temp + self.__d

    def mean_soc_dependency(self, msoc):
        # degradation depends on msoc as a cubic function
        return self.__e * msoc**3 + self.__f * msoc**2 + self.__g * msoc + self.__h

    def doc_dependency(self, doc):
        # degradation depends on doc as an exponential function
        return np.exp(self.__i * doc)

    def crate_dependency(self, crate):
        # degradation depends on crate as an exponential function
        return np.exp(self.__m * crate)

    def calculate_degradation(self, battery_state: LithiumIonState) -> None:
        c_rate: float = self._cycle_detector.get_crate() * 3600
        depth_of_discharge: float = self._cycle_detector.get_depth_of_cycle()
        full_equivalent_cycle: float = self._cycle_detector.get_full_equivalent_cycle()
        mean_soc: float = self._cycle_detector.get_mean_soc()
        temperature: float = battery_state.temperature
        delta_fec: float = self._cycle_detector.get_delta_full_equivalent_cycle()

        # calculate degradation in percent (divide by 100 to get p.u.) with function 26 from MA Sandner
        capacity_loss_after = self.fec_dependency(full_equivalent_cycle) * \
                              self.temperature_dependency(temperature) * \
                              self.mean_soc_dependency(mean_soc) * \
                              self.doc_dependency(depth_of_discharge) * \
                              self.crate_dependency(c_rate)
        capacity_loss_before = self.fec_dependency(full_equivalent_cycle - delta_fec) * \
                               self.temperature_dependency(temperature) * \
                               self.mean_soc_dependency(mean_soc) * \
                               self.doc_dependency(depth_of_discharge) * \
                               self.crate_dependency(c_rate)

        self.__capacity_loss = (capacity_loss_after - capacity_loss_before) / 100.0 * self.__initial_capacity

    def calculate_resistance_increase(self, battery_state: LithiumIonState) -> None:
        # no resistance increase calculated for generic degradation model
        pass

    def get_degradation(self) -> float:
        degradation = self.__capacity_loss
        self.__capacity_loss = 0
        return degradation

    def get_resistance_increase(self) -> float:
        # no resistance increase calculated for generic degradation model
        return 0

    def reset(self, lithium_ion_state: LithiumIonState) -> None:
        self.__capacity_loss = 0

    def close(self) -> None:
        pass



from simses.commons.profile.power.generation import GenerationProfile
from simses.commons.state.energy_management import EnergyManagementState
from simses.commons.state.system import SystemState
from simses.logic.energy_management.strategy.operation_priority import OperationPriority
from simses.logic.energy_management.strategy.operation_strategy import OperationStrategy


class UseAllRenewableEnergy(OperationStrategy):
    """
    This operation strategy is for a plant that uses the whole energy provided by a renewable energy source.
    This strategy is implemented especially for an electrolyzer which produces hydrogen with the energy out
    of a solar or a wind energy plant.
    """
    def __init__(self, pv_generation_profile: GenerationProfile):
        super().__init__(OperationPriority.MEDIUM)
        self.__pv_profile: GenerationProfile = pv_generation_profile
        self.__pv_power = 0
        self.__load_power = 0

    def next(self, time: float, system_state: SystemState, power: float = 0) -> float:
        self.__pv_power = self.__pv_profile.next(time)
        return self.__pv_power

    def update(self, energy_management_state: EnergyManagementState) -> None:
        energy_management_state.pv_power = self.__pv_power
        energy_management_state.load_power = self.__load_power

    def clear(self) -> None:
        self.__pv_power = 0.0
        self.__load_power = 0.0

    def close(self) -> None:
        self.__pv_profile.close()


import math
from numpy import sign
from simses.system.power_electronics.acdc_converter.abstract_acdc_converter import AcDcConverter


class AcDcConverterIdenticalStacked(AcDcConverter):

    def __init__(self, number_converters: int, switch_value: float, acdc_converter: AcDcConverter, power_electronics_config=None):
        super().__init__(acdc_converter.max_power)
        self.__NUMBER_CONVERTERS: int = number_converters
        self.__SWITCH_VALUE: float = switch_value  # ratio power over rated power
        self.__MAX_POWER_INDIVIDUAL: float = acdc_converter.max_power / number_converters
        self.__converters: [AcDcConverter] = list()
        for converter in range(number_converters):
            self.__converters.append(acdc_converter.create_instance(self.__MAX_POWER_INDIVIDUAL, power_electronics_config))

    def to_ac(self, power: float, voltage: float) -> float:
        if power >= 0.0:
            return 0.0
        power_individual = iter(self.__power_splitter(power))
        power_individual_dc: [float] = list()
        for converter in self.__converters:
            converter: AcDcConverter = converter
            power_individual_dc.append(converter.to_ac(next(power_individual), voltage))
        return sum(power_individual_dc)

    def to_dc(self, power: float, voltage: float) -> float:
        if power <= 0.0:
            return 0.0
        power_individual = iter(self.__power_splitter(power))
        power_individual_dc: [float] = list()
        for converter in self.__converters:
            converter: AcDcConverter = converter
            power_individual_dc.append(converter.to_dc(next(power_individual), voltage))
        return sum(power_individual_dc)

    def to_dc_reverse(self, dc_power: float, voltage: float) -> float:
        if dc_power <= 0.0:
            return 0.0
        power_individual = iter(self.__power_splitter(dc_power))
        power_individual_dc: [float] = list()
        for converter in self.__converters:
            converter: AcDcConverter = converter
            power_individual_dc.append(converter.to_dc_reverse(next(power_individual), voltage))
        return sum(power_individual_dc)

    def to_ac_reverse(self, dc_power: float, voltage: float) -> float:
        if dc_power >= 0:
            return 0.0
        power_individual = iter(self.__power_splitter(dc_power))
        power_individual_dc: [float] = list()
        for converter in self.__converters:
            converter: AcDcConverter = converter
            power_individual_dc.append(converter.to_ac_reverse(next(power_individual), voltage))
        return sum(power_individual_dc)

    def __power_splitter(self, power: float) -> list:
        switch_value = self.__SWITCH_VALUE
        power_individual: [float] = [0.0] * self.__NUMBER_CONVERTERS
        if abs(power) > self.max_power:
            for converter in range(self.__NUMBER_CONVERTERS):
                power_individual[converter] = self.__MAX_POWER_INDIVIDUAL * sign(power)
        elif abs(power) > self.max_power * switch_value:
            equal_distributed_power: float = abs(power) / self.__NUMBER_CONVERTERS
            for converter in range(self.__NUMBER_CONVERTERS):
                power_individual[converter] = max(0.0, min(equal_distributed_power, self.__MAX_POWER_INDIVIDUAL)) * sign(power)
        else:
            remaining_power: float = abs(power)
            max_individual_power: float = self.__MAX_POWER_INDIVIDUAL * switch_value
            number_converters_activated: int = min(math.ceil(remaining_power / max_individual_power), self.__NUMBER_CONVERTERS)
            equal_distributed_power: float = remaining_power / number_converters_activated
            for converter in range(self.__NUMBER_CONVERTERS):
                individual_power: float = max(0.0, min(equal_distributed_power, max_individual_power, remaining_power))
                power_individual[converter] = individual_power * sign(power)
                remaining_power -= individual_power
        return power_individual

    @property
    def volume(self) -> float:
        volume: float = 0.0
        for converter in self.__converters:
            volume += converter.volume
        return volume

    @property
    def mass(self):
        mass: float = 0.0
        for converter in self.__converters:
            mass += converter.mass
        return mass

    @property
    def surface_area(self) -> float:
        surface_area: float = 0.0
        for converter in self.__converters:
            surface_area += converter.surface_area
        return surface_area

    def close(self) -> None:
        for converter in self.__converters:
            converter.close()

    @classmethod
    def create_instance(cls, max_power: float, power_electronics_config=None):
        pass




import math
from datetime import datetime, timedelta

import numpy
from pytz import timezone

from simses.analysis.data.energy_management import EnergyManagementData
from simses.analysis.data.system import SystemData
from simses.analysis.evaluation.economic.revenue_stream.abstract_revenue_stream import RevenueStream
from simses.analysis.evaluation.result import EvaluationResult, Description, Unit
from simses.analysis.utils import get_fractional_years
from simses.commons.config.analysis.economic import EconomicAnalysisConfig


class OperationAndMaintenanceRevenue(RevenueStream):

    """ Calculates the yearly costs due to maintenance and operation of the storage technology"""

    __UTC: timezone = timezone('UTC')
    __BERLIN: timezone = timezone('Europe/Berlin')

    def __init__(self, energy_management_data: EnergyManagementData, system_data: SystemData, economic_analysis_config: EconomicAnalysisConfig):
        super().__init__(energy_management_data, system_data, economic_analysis_config)
        self.__annual_relative_o_and_m_costs = economic_analysis_config.annual_realative_o_and_m_costs  # p.u.
        self.__cashflow_list = []
        self.__shorter_than_one_year = True
        #self.__annual_absolut_o_and_m_costs = self.__annual_relative_o_and_m_costs * self._investment_cost  # EUR
        # self.__investment_cost = self._investment_cost

    def get_cashflow(self) -> numpy.ndarray:
        annual_absolute_o_and_m_cost = self.__annual_relative_o_and_m_costs * self._investment_cost
        time = self._energy_management_data.time  # UTC+0

        om_cost_per_timestep = annual_absolute_o_and_m_cost / self._seconds_per_year * (time[1] - time[0])
        om_cost_time_series = [(-1) * om_cost_per_timestep] * len(time)

        # create cashflow list for each project year
        self.__cashflow_list = self.cash_time_series_to_project_years(om_cost_time_series, time)
        return numpy.array(self.__cashflow_list)

    def get_evaluation_results(self) -> [EvaluationResult]:
        key_results: [EvaluationResult] = list()
        key_results.append(EvaluationResult(Description.Economical.OperationAndMaintenance.O_AND_M_COST,
                                            Unit.EURO, [-1 * i for i in self.__cashflow_list]))
        return key_results

    def get_assumptions(self) -> [EvaluationResult]:
        assumptions: [EvaluationResult] = list()
        assumptions.append(EvaluationResult(Description.Economical.OperationAndMaintenance.ANNUAl_O_AND_M_COST,
                                            Unit.EURO, self.__annual_relative_o_and_m_costs * self._investment_cost))
        return assumptions

    def close(self):
        pass

